(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * UPF
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: Analysis.thy 9476 2012-10-15 15:08:28Z brucker $ *)


header{* Algebraic Properties of Policies. *}
theory Analysis
imports
 ParallelComposition
 SeqComposition
begin


text {* In this theory, several standard policy properties are paraphrased in UPF terms. *}

text{* A policy has no gaps *}

definition gap_free :: "('a \<mapsto> 'b) \<Rightarrow> bool" where
 "gap_free p = (dom p = UNIV)"

text{* Policy p is more defined than q *}

definition more_defined :: "('a \<mapsto> 'b) \<Rightarrow>('a \<mapsto> 'b) \<Rightarrow>bool" where
 "more_defined p q = (dom q \<subset> dom p)"

text{* p is more permissive than q *}

definition more_permissive :: "('a \<mapsto> 'b) \<Rightarrow> ('a \<mapsto> 'b) \<Rightarrow> bool" where
 "more_permissive p q = (\<forall> x. (case q x of \<lfloor>allow y\<rfloor> \<Rightarrow> (\<exists> z. (p x = \<lfloor>allow z\<rfloor>))
                                         | \<lfloor>deny y\<rfloor> \<Rightarrow> True
                                         | \<bottom> \<Rightarrow> True))" 

definition more_rejective :: "('a \<mapsto> 'b) \<Rightarrow> ('a \<mapsto> 'b) \<Rightarrow> bool" where
 "more_rejective p q = (\<forall> x. (case q x of \<lfloor>deny y\<rfloor> \<Rightarrow> (\<exists> z. (p x = \<lfloor>deny z\<rfloor>))
                                         | \<lfloor>allow y\<rfloor> \<Rightarrow> True
                                         | \<bottom> \<Rightarrow> True))" 


 lemma "more_permissive (A\<^sub>f f) p"
apply (simp add: more_permissive_def allow_all_fun_def allow_pfun_def)
apply (rule allI)
apply (case_tac "p x" ,simp_all)
apply (case_tac a,simp_all)
done


lemma "more_permissive A\<^sub>I p"
apply (simp add: more_permissive_def allow_all_fun_def allow_pfun_def allow_all_id_def)
apply (rule allI)
apply (case_tac "p x" ,simp_all)
apply (case_tac a,simp_all)
done



text{* Equivalence over domain D *}

definition p_eq_dom :: "('a \<mapsto> 'b) \<Rightarrow>('a \<mapsto> 'b) \<Rightarrow>'a set \<Rightarrow>bool" where
 "p_eq_dom p q d = (\<forall>x\<in>d. p x = q x)"

text{* p and q have no conflicts *}
definition no_conflicts :: "('a \<mapsto> 'b) \<Rightarrow>('a \<mapsto> 'b)  \<Rightarrow>bool" where
 "no_conflicts p q = (dom p = dom q \<and> (\<forall>x\<in>(dom p). 
    (case p x of \<lfloor>allow y\<rfloor> \<Rightarrow> (\<exists>z. q x = \<lfloor>allow z\<rfloor>)
               | \<lfloor>deny y\<rfloor> \<Rightarrow> (\<exists>z. q x = \<lfloor>deny z\<rfloor>))))"


lemma policy_eq:
  assumes p_over_qA: "more_permissive p q"
  and  q_over_pA: "more_permissive q p" 
  and  p_over_qD: "more_rejective q p" 
  and  q_over_pD: "more_rejective p q" 
  and  dom_eq: "dom p = dom q"
 shows "no_conflicts p q"
apply (insert p_over_qA q_over_pA p_over_qD q_over_pD dom_eq)
apply (simp add:  no_conflicts_def more_permissive_def more_rejective_def
          split: option.splits decision.splits)
apply safe
apply (metis domI domIff dom_eq)
apply (metis)+
done

lemma dom_inter: "\<lbrakk>dom p \<inter> dom q = {}; p x = \<lfloor>y\<rfloor>\<rbrakk> \<Longrightarrow> q x = \<bottom>"
by auto

lemma dom_eq: "dom p \<inter> dom q = {} \<Longrightarrow>
  p \<Oplus>\<^sub>A q = p \<Oplus>\<^sub>D q"
apply (simp add: override_A_def override_D_def)
apply (rule ext)
apply (simp split: prod.splits option.splits decision.splits )
apply (simp add: dom_def)
apply auto
done



lemma dom_split_alt_def :
  "(f, g) \<Delta> p = (dom(p \<triangleright> Allow) \<triangleleft> (A\<^sub>f f)) \<Oplus>
                 (dom(p \<triangleright> Deny) \<triangleleft> (D\<^sub>f g))"
apply(rule ext)
apply(simp add: dom_split2_def Allow_def Deny_def dom_restrict_def
                deny_all_fun_def allow_all_fun_def map_add_def)
apply(simp split:option.splits decision.splits)
apply(auto simp:map_add_def o_def deny_pfun_def ran_restrict_def image_def)
done


end
