(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * UPF
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: Normalisation.thy 9476 2012-10-15 15:08:28Z brucker $ *)

header{* Policy Transformations *}
theory Normalisation
imports ParallelComposition
begin


text{* This theory provides the formalisations required for the
transformation of UPF policies. A typical usage scenario can be
observed in the firewall case study.
*}


subsection{* Elementary Operators *}

text{* We start by providing several operatos and theorems useful when reasoning about 
       a list of rules which should eventually be interpreted as combined using the 
       standard override operator. *}


text{* The following definition takes as argument a list of rules and returns a policy where the
       rules are combined using the standard override operator.*}
definition list2policy::"('a \<mapsto> 'b) list \<Rightarrow>
                          ('a \<mapsto> 'b)"  where
 "list2policy l = (foldr (\<lambda> x y. (x \<Oplus> y)) l \<emptyset>)"


text{* Determine the position of element of a list. *}
fun position :: "'\<alpha> \<Rightarrow> '\<alpha> list \<Rightarrow>  nat" where
  "position a [] = 0"
|"(position a (x#xs)) = (if a = x then 1 else (Suc (position a xs)))" 


text{* Provides the first applied rule of a policy given as a list of rules. *}
fun  applied_rule where
"applied_rule C a (x#xs) = (if a \<in> dom (C x) then (Some x)
                               else (applied_rule C a xs))"
|"applied_rule C a [] = None"

text {* The following is used if the list is constructed backwards. *}
definition applied_rule_rev where
"applied_rule_rev C a x =  (applied_rule C a (rev x))"


text{* The following is a typical policy transformation. It can be applied to any type of policy and
      removes all the rules from a policy with an empty domain. It takes two arguments: a semantic
      interpretation function and a list of rules. *}
fun removeShadowRules3  where
 "removeShadowRules3 C (x#xs) = (if (dom (C x)= {}) then (removeShadowRules3 C xs)
                               else (x#(removeShadowRules3 C xs)))"
|"removeShadowRules3 C [] = []"



text {* The following invariant establishes that there are no rules with an empty domain
        in a list of rules. *}

fun noneMT where 
 "noneMT C (x#xs) = (dom (C x) \<noteq> {} \<and>  (noneMT C xs))"
|"noneMT C [] = True" 


text{* The following related invariant establishes that the policy has not a 
    completely empty domain. *}

fun notMTpolicy where
 "notMTpolicy C (x#xs) = (if (dom (C x) = {}) then (notMTpolicy C xs) else True)"
|"notMTpolicy C [] = False"


text{* Next, a few theorems about the two invariants and the transformation*}

lemma noneMT_vs_notMT[rule_format]: "noneMT  C p \<longrightarrow> p \<noteq> [] \<longrightarrow> notMTpolicy C p" 
apply (induct p)
apply simp_all
done


lemma RS3nMT[rule_format]: "noneMT C  (removeShadowRules3 C p)"
by (induct p,simp_all)


lemma RS3nMT2[rule_format]: "noneMT C  p \<Longrightarrow>  (removeShadowRules3 C p) = p"
by (induct p,simp_all)



lemma nMTcharn: "noneMT C p = (\<forall> r \<in> set p. dom (C r) \<noteq> {})"
by (induct p) simp_all

lemma nMTeqSet: "set p = set s \<Longrightarrow> noneMT C p = noneMT C s"
by (simp add: nMTcharn)


lemma notMTnMT: "\<lbrakk>a \<in> set p; noneMT C p\<rbrakk> \<Longrightarrow> dom (C a) \<noteq> {}"
by (simp add: nMTcharn)

lemma noneMTconc[rule_format]: "noneMT C (a@[b]) \<longrightarrow> noneMT C a"
by (induct a, simp_all)

lemma nMTtail[rule_format]: "noneMT C p \<longrightarrow> noneMT C (tl p)"
by (induct p, simp_all)


lemma notMTpolicyimpnotMT[simp]: "notMTpolicy C p \<Longrightarrow> p \<noteq> []"
by auto

lemma SR3nMT[rule_format]: "\<not> notMTpolicy C  p \<longrightarrow> removeShadowRules3 C p = []"
by (induct p, simp_all)


lemma NMPcharn[rule_format]: "a \<in> set p\<longrightarrow> dom (C a) \<noteq> {} \<longrightarrow> notMTpolicy C  p"
by (induct p, simp_all)


lemma NMPRS3[rule_format]: "notMTpolicy C  p \<longrightarrow> notMTpolicy C  (removeShadowRules3 C p)"
by (induct p, simp_all)


text{* Next, a few theorems about applied\_rule *}

lemma mrconc[rule_format]: "applied_rule_rev C x p = Some a \<longrightarrow>
                            applied_rule_rev C x (b#p) = Some a"
apply (rule rev_induct)  back 
apply (simp_all add: applied_rule_rev_def)
done

lemma mreq_end: "\<lbrakk>applied_rule_rev C x b = Some r; applied_rule_rev C x c = Some r\<rbrakk> \<Longrightarrow> 
 applied_rule_rev C x (a#b) = applied_rule_rev C x (a#c)"
by (simp add: mrconc)

lemma mrconcNone[rule_format]: "applied_rule_rev C x p = None \<longrightarrow>
                                applied_rule_rev C x (b#p) = applied_rule_rev C x [b]"
apply (rule_tac xs = p in rev_induct) 
apply simp_all
apply (rule impI)
apply (case_tac "x \<in> dom (C xa)")
apply (simp_all add: applied_rule_rev_def)
done

lemma mreq_endNone: "\<lbrakk>applied_rule_rev C x b = None; applied_rule_rev C x c = None\<rbrakk> \<Longrightarrow> 
     applied_rule_rev C x (a#b) = applied_rule_rev C x (a#c)"
by (metis mrconcNone)

lemma mreq_end2: "applied_rule_rev C x b = applied_rule_rev C x c \<Longrightarrow> 
     applied_rule_rev C x (a#b) = applied_rule_rev C x (a#c)"
apply (case_tac "applied_rule_rev C x b = None")
apply (auto intro: mreq_end mreq_endNone)
done

lemma mreq_end3: "applied_rule_rev C x p \<noteq> None \<Longrightarrow>
                  applied_rule_rev C x (b # p) = applied_rule_rev C x (p)"
by (auto simp: mrconc)

lemma mrNoneMT[rule_format]: "r \<in> set p \<longrightarrow> applied_rule_rev C x p = None \<longrightarrow>
                              x \<notin> dom (C r)"
apply (rule rev_induct, simp_all)
apply (rule conjI| rule impI)+
apply simp_all
apply (case_tac "xa \<in> set xs")
apply (simp_all add: applied_rule_rev_def split: if_splits)
done





subsection{* Distributivity of the Transformation. *}

text{*
The scenario is the following (can be applied iteratively):
 - Two policies are combined using one of the parallel combinators
 - (e.g. P = P1 P2) (At least) one of the constituent policies has
 - a normalisation procedures, which as output produces a list of
 - policies that are semantically equivalent to the original policy if
 - combined from left to right using the override operator.
*}

text{* The following function is crucial for the distribution. Its arguments are a policy, a list
       of policies, a parallel combinator, and a range and a domain coercion function. *}

fun prod_list :: "('\<alpha> \<mapsto>'\<beta>) \<Rightarrow> (('\<gamma> \<mapsto>'\<delta>) list) \<Rightarrow> 
                  (('\<alpha> \<mapsto>'\<beta>) \<Rightarrow> ('\<gamma> \<mapsto>'\<delta>) \<Rightarrow> (('\<alpha> \<times> '\<gamma>) \<mapsto> ('\<beta> \<times> '\<delta>))) \<Rightarrow>
                  (('\<beta> \<times> '\<delta>) \<Rightarrow> 'y) \<Rightarrow> ('x \<Rightarrow> ('\<alpha> \<times> '\<gamma>)) \<Rightarrow>  
                  (('x \<mapsto> 'y) list)"  (infixr "\<Otimes>\<^sub>L" 54) where
  "prod_list x (y#ys) par_comb ran_adapt dom_adapt = 
  ((ran_adapt o_f ((par_comb x y) o dom_adapt))#(prod_list x ys par_comb ran_adapt dom_adapt))"
| "prod_list x [] par_comb ran_adapt dom_adapt = []"


text{* An instance, as usual there are four of them. *}

definition prod_2_list :: "[('\<alpha> \<mapsto>'\<beta>), (('\<gamma> \<mapsto>'\<delta>) list)] \<Rightarrow> 
                  (('\<beta> \<times> '\<delta>) \<Rightarrow> 'y) \<Rightarrow> ('x \<Rightarrow> ('\<alpha> \<times> '\<gamma>)) \<Rightarrow> 
                  (('x \<mapsto> 'y) list)" (infixr "\<Otimes>\<^sub>2\<^sub>L" 55) where 
  "x \<Otimes>\<^sub>2\<^sub>L y =  (\<lambda> d r. (x \<Otimes>\<^sub>L y) (op \<Otimes>\<^sub>2) d r)"  



lemma list2listNMT[rule_format]:  "x \<noteq> [] \<longrightarrow> map sem x \<noteq> []"
apply (case_tac x)
apply simp_all
done
 
lemma two_conc: "(prod_list x (y#ys) p r d)=  
((r o_f ((p x y) o d))#(prod_list x ys p r d))"
by simp



text{* The following two invariants establish if the law of distributivity holds for a combinator
       and if an operator is strict regarding undefinedness. *}

definition is_distr where
 "is_distr p = (\<lambda> g f. (\<forall> N P1 P2. ((g o_f ((p N (P1 \<Oplus> P2)) o f)) = 
               ((g o_f ((p N P1) o f)) \<Oplus> (g o_f ((p N P2)  o f))))))"



definition is_strict where
 "is_strict p = (\<lambda> r d. \<forall> P1. (r o_f (p P1 \<emptyset> \<circ> d)) = \<emptyset>)"


lemma is_distr_orD: "is_distr (op \<Otimes>\<^sub>\<or>\<^sub>D) d r"
apply (simp add: is_distr_def)
apply (rule allI)+
apply (rule distr_orD)
by simp


lemma is_strict_orD: "is_strict (op \<Otimes>\<^sub>\<or>\<^sub>D) d r"
apply (simp add: is_strict_def)
apply (simp add: policy_range_comp_def)
done


lemma is_distr_2: "is_distr (op \<Otimes>\<^sub>2) d r"
apply (simp add: is_distr_def)
apply (rule allI)+
apply (rule distr_or2)
by simp

lemma is_strict_2: "is_strict (op \<Otimes>\<^sub>2) d r"
apply (simp only: is_strict_def)
apply simp
apply (simp add: policy_range_comp_def)
done


lemma domStart: "t \<in> dom p1 \<Longrightarrow> (p1 \<Oplus> p2) t = p1 t"
apply (simp add: map_add_dom_app_simps)
done


lemma notDom: "x \<in> dom A \<Longrightarrow> \<not> A x = None"
apply auto
done


declare Fun.o_apply [simp del]

text{* The following theorems help to establish the main correctness theorem of the distribution.
       They are provided for all four parallel combination operators. *}

lemma distr_aux1[rule_format]: 
 "x \<in> dom (r o_f (A \<circ> d)) \<longrightarrow>
    (r o_f  (A \<Oplus> B) o d) x = (r o_f (A \<circ> d)) x"
by (simp add: prod_2_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)



lemma distr_aux2_1[rule_format]: "x \<in> dom (r o_f ((P1 \<Otimes>\<^sub>2 a) \<circ> d)) \<longrightarrow>
         ((r o_f ((P1 \<Otimes>\<^sub>2 a \<Oplus> Q) \<circ> d)) x = (r o_f ((P1 \<Otimes>\<^sub>2 a) \<circ> d)) x) "
apply (simp add: prod_2_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done



lemma distr_aux2_2[rule_format]: 
 "x \<notin> dom (r o_f ((P1 \<Otimes>\<^sub>2 a) \<circ> d)) \<longrightarrow>
  (((r o_f ((P1 \<Otimes>\<^sub>2 a \<Oplus> b) \<circ> d)) x) =
  ((r o_f  ((P1 \<Otimes>\<^sub>2 b) \<circ> d)) x))"
apply (simp add: prod_2_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done


lemma distr_aux1_1[rule_format]: "x \<in> dom (r o_f ((P1 \<Otimes>\<^sub>1 a) \<circ> d)) \<longrightarrow>
         ((r o_f ((P1 \<Otimes>\<^sub>1 a \<Oplus> Q) \<circ> d)) x = (r o_f ((P1 \<Otimes>\<^sub>1 a) \<circ> d)) x) "
apply (simp add: prod_1_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done


lemma 
 "x \<notin> dom ( (P1 \<Otimes>\<^sub>1 a) \<circ> d) \<longrightarrow>
  (((((P1 \<Otimes>\<^sub>1 a \<Oplus> b) \<circ> d)) x) =
  ((((P1 \<Otimes>\<^sub>1 b) \<circ> d)) x))"
apply (simp add: prod_1_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done

lemma distr_aux1_2[rule_format]: 
 "x \<notin> dom (r o_f ((P1 \<Otimes>\<^sub>1 a) \<circ> d)) \<longrightarrow>
  (((r o_f ((P1 \<Otimes>\<^sub>1 a \<Oplus> b) \<circ> d)) x) =
  ((r o_f  ((P1 \<Otimes>\<^sub>1 b) \<circ> d)) x))"
apply (simp add: prod_1_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done


lemma distr_auxA_1[rule_format]: "x \<in> dom (r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>A a) o d)) \<longrightarrow>
         ((r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>A a \<Oplus> Q) \<circ> d)) x = (r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>A a) \<circ> d)) x) "
apply (simp add: prod_orA_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done


lemma 
 "x \<notin> dom ( (P1 \<Otimes>\<^sub>\<or>\<^sub>A a) \<circ> d) \<longrightarrow>
  (((((P1 \<Otimes>\<^sub>\<or>\<^sub>A a \<Oplus> b) \<circ> d)) x) =
  ((((P1 \<Otimes>\<^sub>\<or>\<^sub>A b) \<circ> d)) x))"
apply (simp add: prod_orA_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done

lemma distr_auxA_2[rule_format]: 
 "x \<notin> dom (r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>A a) \<circ> d)) \<longrightarrow>
  (((r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>A a \<Oplus> b) \<circ> d)) x) =
  ((r o_f  ((P1 \<Otimes>\<^sub>\<or>\<^sub>A b) \<circ> d)) x))"
apply (simp add: prod_orA_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done


lemma distr_auxD_1[rule_format]: "x \<in> dom (r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>D a) \<circ> d)) \<longrightarrow>
         ((r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>D a \<Oplus> Q) \<circ> d)) x = (r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>D a) \<circ> d)) x) "
apply (simp add: prod_orD_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done


lemma  
 "x \<notin> dom ( (P1 \<Otimes>\<^sub>\<or>\<^sub>D a) \<circ> d) \<longrightarrow>
  (((((P1 \<Otimes>\<^sub>\<or>\<^sub>D a \<Oplus> b) \<circ> d)) x) =
  ((((P1 \<Otimes>\<^sub>\<or>\<^sub>D b) \<circ> d)) x))"
apply (simp add: prod_orD_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done

lemma distr_auxD_2[rule_format]: 
 "x \<notin> dom (r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>D a) \<circ> d)) \<longrightarrow>
  (((r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>D a \<Oplus> b) \<circ> d)) x) =
  ((r o_f  ((P1 \<Otimes>\<^sub>\<or>\<^sub>D b) \<circ> d)) x))"
apply (simp add: prod_orD_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done


text{* The following theorems are crucial: they establish the correctness of the distribution.
        *}
lemma Norm_Distr_1[rule_format]: 
 "((r o_f (((op \<Otimes>\<^sub>1) P1 (list2policy P2)) o d)) x = 
  ((list2policy ((P1 \<Otimes>\<^sub>L P2) (op \<Otimes>\<^sub>1) r d)) x))"
apply (induct P2) 
apply (simp add: policy_range_comp_def o_apply list2policy_def)
apply (simp add: list2policy_def)
apply (case_tac "x \<in> dom (r o_f ((P1 \<Otimes>\<^sub>1 a) \<circ> d))") 
apply (auto simp: domStart distr_aux1_1  distr_aux1_2 list2policy_def map_add_dom_app_simps(3) prod_2_list_def)
done

lemma Norm_Distr_2[rule_format]: 
 "((r o_f (((op \<Otimes>\<^sub>2) P1 (list2policy P2)) o d)) x = 
  ((list2policy ((P1 \<Otimes>\<^sub>L P2) (op \<Otimes>\<^sub>2) r d)) x))"
apply (induct P2) 
apply (simp add: policy_range_comp_def o_apply list2policy_def)
apply (simp add: list2policy_def)
apply (case_tac "x \<in> dom (r o_f ((P1 \<Otimes>\<^sub>2 a) \<circ> d))") 
apply (auto simp: domStart distr_aux2_1  distr_aux2_2 list2policy_def map_add_dom_app_simps(3) prod_2_list_def)
done

lemma Norm_Distr_A[rule_format]: 
 "((r o_f (((op \<Otimes>\<^sub>\<or>\<^sub>A) P1 (list2policy P2)) o d)) x = 
  ((list2policy ((P1 \<Otimes>\<^sub>L P2) (op \<Otimes>\<^sub>\<or>\<^sub>A) r d)) x))"
apply (induct P2) 
apply (simp add: policy_range_comp_def o_apply list2policy_def)
apply (simp add: list2policy_def)
apply (case_tac "x \<in> dom (r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>A a) \<circ> d))") 
apply (auto simp: domStart distr_auxA_1 distr_auxA_2 list2policy_def map_add_dom_app_simps(3) prod_2_list_def)
done

lemma Norm_Distr_D[rule_format]: 
 "((r o_f (((op \<Otimes>\<^sub>\<or>\<^sub>D) P1 (list2policy P2)) o d)) x = 
  ((list2policy ((P1 \<Otimes>\<^sub>L P2) (op \<Otimes>\<^sub>\<or>\<^sub>D) r d)) x))"
apply (induct P2) 
apply (simp add: policy_range_comp_def o_apply list2policy_def)
apply (simp add: list2policy_def)
apply (case_tac "x \<in> dom (r o_f ((P1 \<Otimes>\<^sub>\<or>\<^sub>D a) \<circ> d))") 
apply (auto simp: domStart distr_auxD_1 distr_auxD_2 list2policy_def map_add_dom_app_simps(3) prod_2_list_def)
done


text {* Some domain reasoning *}

lemma domSubsetDistr1: "dom A = UNIV \<Longrightarrow> dom ((\<lambda>(x, y). x) o_f (A \<Otimes>\<^sub>1 B) o (\<lambda> x. (x,x))) = dom B"
apply (rule set_eqI)
apply (rule iffI)
by (auto simp: prod_1_def policy_range_comp_def dom_def o_apply
              split: decision.splits option.splits prod.splits)


lemma domSubsetDistr2: "dom A = UNIV \<Longrightarrow> dom ((\<lambda>(x, y). x) o_f (A \<Otimes>\<^sub>2 B) o (\<lambda> x. (x,x))) = dom B"
apply (rule set_eqI)
apply (rule iffI)
by (auto simp: prod_2_def policy_range_comp_def dom_def o_apply
              split: decision.splits option.splits prod.splits)


lemma domSubsetDistrA: "dom A = UNIV \<Longrightarrow> dom ((\<lambda>(x, y). x) o_f (A \<Otimes>\<^sub>\<or>\<^sub>A B) o (\<lambda> x. (x,x))) = dom B"
apply (rule set_eqI)
apply (rule iffI)
by (auto simp: prod_orA_def policy_range_comp_def dom_def o_apply
              split: decision.splits option.splits prod.splits)


lemma domSubsetDistrD: "dom A = UNIV \<Longrightarrow> dom ((\<lambda>(x, y). x) o_f (A \<Otimes>\<^sub>\<or>\<^sub>D B) o (\<lambda> x. (x,x))) = dom B"
apply (rule set_eqI)
apply (rule iffI)
by (auto simp: prod_orD_def policy_range_comp_def dom_def o_apply
              split: decision.splits option.splits prod.splits)

declare o_apply [simp]


end


