(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * FTPTestDocument.thy --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France*
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: FTPObserver2Document.thy 9225 2011-08-28 21:23:47Z lukasbru $ *)

theory FTPObserver2Document
imports FWTesting
begin

text{* In this theory, we formalise an adapted version of an FTP
protocol using the observers. The protocol consists of four messages:
\begin{itemize} 
 \item portReq X: the client initiates a session, and specifies a port range, where the data should be sent to (only an upper bound, for the sake of simplicity).
 \item portAck Y: the server acknowledges the connection, and non-deterministically chooses a port number from the specified range.
 \item data: the server sends data on the specified port. This message can happen arbitrarily many times.
 \item close: the client closes the connection.
\end{itemize}

We will make use of the observer2, and closely follow the corresponding example from the HOL-TestGen distribution.
*}

text{* The test case generation is done on the basis of \emph{abstract
traces}. Such abstract traces contain explicit variables, and the
functions substitute and rebind are used to replace them with concrete
values during the run of the test driver.*}

datatype vars = X | Y

datatype data = Data

type_synonym chan = port

type_synonym env = "vars \<rightharpoonup> chan"

definition lookup :: "['a \<rightharpoonup> 'b,'a] \<Rightarrow> 'b" where
 "lookup env v = the (env v)"

text{* The traces are lists of packets. However, in this case, we will
not make use of the usual packet definition \emph{directly}, but use a
datatype representation of them. There are abstract and concrete
packets:*}

datatype ftp_packet_abs = port_reqA vars id IntegerPort IntegerPort | 
                          port_ackA vars id IntegerPort IntegerPort | 
                          dataA vars id IntegerPort address | 
                          closeA id IntegerPort IntegerPort

datatype ftp_packet_conc = port_reqC port id IntegerPort IntegerPort | 
                           port_ackC port id IntegerPort IntegerPort | 
                           dataC id IntegerPort IntegerPort | 
                           closeC id IntegerPort IntegerPort


type_synonym ftp_packet = "ftp_packet_abs + ftp_packet_conc"

text{* The following two functions then make the connection between
the packet representations. Note that in the way this function is
defined, a data message will always be allowed. In contrast to the
other form of FTP testing, we do not change the policy during protocol
execution, rather we take more control of the protocol execution
itself: *}

datatype ftp_event = port_req | port_ack | data | close

fun packet_accept :: "ftp_packet_abs \<Rightarrow> (IntegerPort,ftp_event) FWPolicy \<Rightarrow> bool"
where
 "packet_accept (port_reqA v i s d) p = p_accept (i,s,d,port_req) p"
|"packet_accept (port_ackA v i s d) p = p_accept (i,s,d,port_ack) p"
|"packet_accept (closeA  i s d) p = p_accept (i,s,d,close) p"
|"packet_accept (dataA v i s da) p = True"

fun packet_accept_conc ::"ftp_packet_conc \<Rightarrow>(IntegerPort,ftp_event) FWPolicy \<Rightarrow>bool"
where
 "packet_accept_conc (port_reqC v i s d) p = p_accept (i,s,d,port_req) p"
|"packet_accept_conc (port_ackC v i s d) p = p_accept (i,s,d,port_ack) p"
|"packet_accept_conc (closeC i s d) p = p_accept (i,s,d,close) p"
|"packet_accept_conc (dataC i s da) p = True"


text{* The usual function substitute and rebind: *}

fun substitute :: "[env, ftp_packet_abs] \<Rightarrow> ftp_packet_conc" where
 "substitute env (port_reqA v i s d) = (port_reqC (lookup env v) i s d)"
|"substitute env (port_ackA v i s d) = (port_reqC (lookup env v)i s d)"
|"substitute env (dataA v i s da) = (dataC i s (da,(lookup env v)))"
|"substitute env (closeA i s d) = (closeC i s d)"

fun rebind :: "[env, ftp_packet_conc] \<Rightarrow> env" where
 "rebind env (port_reqC p i s d) = env(X \<mapsto> p)"
|"rebind env (port_ackC p i s d) = env(Y \<mapsto> p)"
|"rebind env (dataC i s d) = env"
|"rebind env (closeC i s d) = env"



text{* The automaton which describes successful executions of the protocol:*}

datatype ftp_states = S0 | S1 | S2 | S3

fun ftp_automaton ::
  "ftp_states \<Rightarrow> ftp_packet_abs list \<Rightarrow> (IntegerPort,ftp_event) FWPolicy \<Rightarrow>
                      id \<Rightarrow> IntegerPort \<Rightarrow> IntegerPort \<Rightarrow> bool" where
 "ftp_automaton H [] p in_id c s  = (H = S3)"
|"ftp_automaton H (x#xs) policy i c s = (case H of  
  S0 => (case x of (port_reqA X in_id sr de) \<Rightarrow> i = in_id \<and> sr = c \<and> de = s \<and>
               packet_accept x policy \<and> ftp_automaton S1 xs policy i c s
                   | _ \<Rightarrow> False)
| S1 => (case x of (port_ackA Y in_id sr de) \<Rightarrow> i = in_id \<and> sr = s \<and> de = (fst c,21) \<and>
                    packet_accept x policy \<and> ftp_automaton S2 xs policy i c s
                    | _ \<Rightarrow> False)
| S2 \<Rightarrow> (case x of (dataA Y in_id sr da) \<Rightarrow> i = in_id \<and> sr = s \<and> fst c = (da) \<and>
                                             ftp_automaton S2 xs policy i c s
                | (closeA in_id sr de) \<Rightarrow> i = in_id \<and> sr = c \<and> de = s \<and>
                     packet_accept x policy \<and> ftp_automaton S3 xs policy i c s
                | _ \<Rightarrow> False)
| S3 \<Rightarrow> False)"

text{* TODO: Alternative, wo die Policy explizit geaendert wird. *}


text{* Next, we declare our specific setting and the policy: *}
definition 
  intranet :: "IntegerPort net" where
  "intranet = {{(a,e) . a = 3}}"

definition
  internet :: "IntegerPort net" where
  "internet = {{(a,c). a > 3}}"


definition 
  ftp_policy :: "(IntegerPort,ftp_event) FWPolicy" where
  "ftp_policy = (deny_all ++ allow_from_to_port (21::port) internet intranet ++
                              allow_all_from_to intranet internet)"


text{* The next two constants check if an address is in the Intranet
or in the Internet respectively.*}

definition 
  is_in_intranet :: "IntegerPort  \<Rightarrow> bool" where
  "is_in_intranet a = ((fst a) = 3)"

definition
  is_in_internet :: "IntegerPort \<Rightarrow> bool" where
  "is_in_internet a =  ((fst a) > 3)"



definition 
  NB_ftp  where
  "NB_ftp i c s = {x. (ftp_automaton S0 x ftp_policy i c s)}"

definition "accept_ftp" ::  "ftp_packet_abs list \<Rightarrow> bool" where
"accept_ftp t = (\<exists> i c s. t \<in> NB_ftp i c s \<and> is_in_intranet c \<and> is_in_internet s)"



text{* The postcondition: *}
fun postcond :: "env \<Rightarrow> '\<sigma> \<Rightarrow> ftp_packet_conc \<Rightarrow> ftp_packet_conc \<Rightarrow> bool" where 
  "postcond env x (port_reqC p i c s) y = (case y of (port_ackC pa i s c) =>
                                                        (pa <= p) | _ \<Rightarrow> False)"
| "postcond env x (port_ackC p i s c) y = (case y of (dataC i s c) =>
                                    (snd c = p \<and> p = lookup env Y) |_ \<Rightarrow> False)"
| "postcond env x (dataC i s c) y = (case y of (dataC i s c) \<Rightarrow>
                                                         (snd c = lookup env Y)
                                              |(closeC i c s) \<Rightarrow> True)"
| "postcond env x y z = False"


lemmas foo_lemmas = NB_ftp_def accept_ftp_def ftp_policy_def accept_ftp_def
 p_accept_def intranet_def internet_def
is_in_intranet_def is_in_internet_def 

declare foo_lemmas[simp,iff]


text{* Next some theorem proving, trying to achieve better test case generation results: *}
lemma allowAll[simp]: "packet_accept x allow_all"
apply (case_tac x, simp_all)
apply (simp_all add: p_accept_def allow_all_def)
done

lemma start[simp]: "ftp_automaton S0 (x#xs) p i c s = ((x = port_reqA X i c s) \<and>
                               ftp_automaton S1 xs p i c s \<and> packet_accept x p)"
apply simp
apply (case_tac x,simp_all)
apply (rule vars.exhaust, auto)
done

lemma step1[simp]: 
 "ftp_automaton S1 (x#xs) p i c s = ((x = port_ackA Y i s (fst c,21)) \<and>
  ftp_automaton S2 xs p i c s \<and> packet_accept x p)"
apply simp
apply (case_tac x,simp_all)
apply (case_tac vars,simp_all)
apply (rule vars.exhaust, auto)
done


lemma step2[simp]: "ftp_automaton S2 (x#xs) p i c s =
 ((x = dataA Y i s (fst c)) \<and> ftp_automaton S2 xs p i c s\<and> packet_accept x p) \<or>
 ((x = closeA i c s) \<and> ftp_automaton S3 xs p i c s)"
apply simp
apply (case_tac x,simp_all)
apply (case_tac vars,simp_all)
apply (rule vars.exhaust, auto)
done



lemma step3[simp]:
"ftp_automaton S2 [x] p i c s = (x = closeA i c s\<and> packet_accept x p)"
apply simp
apply (case_tac x, simp_all)
apply (case_tac vars)
apply simp
apply simp
apply auto
done


lemma packet_accept_a[simp]: "packet_accept (dataA a b c d) p"
apply simp
done

lemma packet_accept_b[simp]: "is_in_intranet c \<and> is_in_internet s \<Longrightarrow>
                             packet_accept (port_reqA x i c s) ftp_policy"
apply simp
apply (simp add: dest_def src_def in_subnet_def IntegerPort.dest_port_def
 PortCombinators packet_accept_def PolicyCombinators intranet_def internet_def)
apply auto
done

lemma packet_accept_c[simp]:"is_in_intranet c \<and> is_in_internet s \<and> snd c = 21\<Longrightarrow>
                            packet_accept (port_ackA y i s c) ftp_policy"
apply simp
apply (simp add: dest_def src_def in_subnet_def IntegerPort.dest_port_def
 PortCombinators packet_accept_def PolicyCombinators intranet_def internet_def)
apply auto
sorry

lemma packet_accept_d[simp]: "is_in_intranet c \<and> is_in_internet s \<Longrightarrow>
                            packet_accept (closeA i c s) ftp_policy"
apply simp
apply (simp add: dest_def src_def in_subnet_def IntegerPort.dest_port_def
 PortCombinators packet_accept_def PolicyCombinators intranet_def internet_def)
apply auto
done

text{* Now the test specification: *}
test_spec "accept_ftp t \<longrightarrow>
       (([X\<mapsto>init_value],()) \<models> (os \<leftarrow>
           (mbind t (observer2 rebind substitute postcond ioprog));
           result (length t = length os)))"
apply (simp add: accept_ftp_def NB_ftp_def accept_ftp_def
                 p_accept_def intranet_def internet_def is_in_intranet_def
                 is_in_internet_def PortCombinators packet_defs IntegerPortLemmas
            split: prod.splits)
apply (gen_test_cases 5 1 "ioprog")
store_test_thm "ftp"

declare [[testgen_SMT=true]]
declare [[testgen_iterations=100]]

declare ftp_event.splits [iff] 
declare packet_accept_def[simp]
declare ftp_packet_abs.splits [iff] 
declare allow_all_from_to_def [simp]
gen_test_data "ftp"

thm ftp.test_data
thm ftp.pos


end