(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * FTPTestDocument.thy --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France*
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: FTPTestDocument.thy 9225 2011-08-28 21:23:47Z lukasbru $ *)

theory FTPTestDocument
imports 
  FWTesting
begin


text{* In this theory we generate the test data for correct runs of
the FTP protocol. As usual, we start with definining the networks and
the policy. We use a rather simple policy which allows only FTP
connections starting from the intranet going to the internet and
denies everything else. *}

definition
  intranet :: "IntegerPort net" where
  "intranet = {{(a,b) . a = 3}}"

definition
  internet :: "IntegerPort net" where
  "internet = {{(a,b). a > 3}}"

definition
  ftp_policy :: "(IntegerPort,ftp_msg) FWPolicy" where
  "ftp_policy = deny_all ++ allow_from_to_port ftp intranet  internet"


text{* The next two constants check if an address is in the Intranet
or in the Internet respectively.*}

definition
  is_in_intranet :: "IntegerPort  \<Rightarrow> bool" where
  "is_in_intranet a = ((fst a) = 3)"

definition
  is_in_internet :: "IntegerPort \<Rightarrow> bool" where
  "is_in_internet a = ((fst a) > 3)"


text{*
The next definition is our starting state: an empty trace and the just defined policy.*}

definition
  "\<sigma>_0_ftp" ::  "(IntegerPort, ftp_msg) history \<times>
                 (IntegerPort, ftp_msg) FWPolicy"    where                                     
  "\<sigma>_0_ftp = ([],ftp_policy)"

text{*Next we state the conditions we have on our trace: a normal
behaviour FTP run from the intranet to some server in the internet
starting on port 21.*}

definition 
"accept_ftp" ::  "(IntegerPort, ftp_msg) history \<Rightarrow> bool" where
          "accept_ftp t = (\<exists> c s i p. t \<in> NB_ftp c s i p \<and> is_in_intranet c \<and>
                                            is_in_internet s \<and> (snd s) = 21)"


text{*The depth of the test case generation corresponds to the maximal
length of generated traces. 4 is the minimum to get a full FTP
protocol run.  *}

declare [[testgen_depth=4]]

text{* The test specification:*}
test_spec "accept_ftp (rev t) \<longrightarrow>
 (\<sigma>_0_ftp \<models> (os  \<leftarrow>  mbind t FTP_ST; (\<lambda> \<sigma>. Some (FUT (rev t) = \<sigma>, \<sigma>))))"
  apply(simp add: accept_ftp_def \<sigma>_0_ftp_def)
  apply (rule impI)+
  apply (unfold NB_ftp_def is_in_internet_def is_in_intranet_def)
  apply simp
  apply (gen_test_cases "FUT" split: HOL.split_if_asm)
  apply (simp_all) 
store_test_thm "ftp_test"


text{* We need to add all required lemmas to the simplifier set, such
that they can be used during test data generation. *}


lemmas ST_simps = Let_def valid_SE_def unit_SE_def bind_SE_def
 in_subnet_def src_def dest_def IntegerPort.dest_port_def
 subnet_of_def id_def port_open_def is_init_def is_data_def
 is_port_request_def is_close_def p_accept_def content_def
 PolicyCombinators PortCombinators is_in_intranet_def
 is_in_internet_def intranet_def internet_def exI subnetOf_lemma
 subnetOf_lemma2 subnetOf_lemma3 subnetOf_lemma4 port_open_lemma
 ftp_policy_def

declare ST_simps [simp]

gen_test_data ftp_test

thm ftp_test.test_data
thm ftp_test.pos

text{*
@{text "gen_test_data ftp_test"}
*}
declare ST_simps [simp del]

text{* The generated test data look as follows (with the unfolded policy rewritten):

\begin{itemize}
\item FUT [(4, (3, 5), (8, 21), $ftp\_close$), (4, (3, 5), (8, 21), $ftp\_port\_request$ 4),
           (4, (3, 5), (8, 21), $ftp\_init$)] =
([(4, (3, 5), (8, 21), $ftp\_close$), (4, (3, 5), (8, 21), $ftp\_port\_request$ 4),
           (4, (3, 5), (8, 21), $ftp\_init$)],policy)
\item FUT [(1, (3, 7), (9, 21), $ftp\_close$), (1, (9, 21), (3, 6), $ftp\_data$), 
      (1, (3, 7), (9, 21), $ftp\_port\_request$ 6),(1, (3, 7), (9, 21), $ftp\_init$)] =
([(1, (3, 7), (9, 21), $ftp\_close$), (1, (9, 21), (3, 6), $ftp\_data$), 
  (1, (3, 7), (9, 21), $ftp\_port\_request$ 6), (1, (3, 7), (9, 21), $ftp\_init$)],policy)
\end{itemize}
*}


end
