(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * List_test.thy --- testing infinite data structures (lists)
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007 ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: List_test.thy 9346 2012-05-03 14:30:04Z wolff $ *)

header {* Testing List Properties *} 

theory 
  List_test 
imports 
  List
  Testing
begin


text {* In this example we present the current main application of
\testgen: generating test data for black box testing of functional
programs within a specification based unit test. We use a simple
scenario, developing the test theory for testing sorting algorithms
over lists.  *}

section {* A Quick Walk Through *} 
text {* In the following we give a
first impression of how the testing process using \testgen{} looks
like. For brevity we stick to default parameters and explain possible
decision points and parameters where the testing can be improved in
the next section.  *}

subsubsection {* Writing the Test Specification *}
text {* We start by specifying a primitive recursive predicate
describing sorted lists: *}

primrec is_sorted:: "('a::linorder) list \<Rightarrow> bool"
  where "is_sorted  []      = True" |
        "is_sorted  (x#xs)  = (case xs of 
                                []   \<Rightarrow> True 
                              | y#ys \<Rightarrow> x \<le> y \<and> is_sorted xs)"


text {*
We will use this HOL predicate for describing our test
specification, i.e. the properties our implementation should
fulfill:
*}
test_spec "is_sorted(PUT (l::('a list)))"
(*<*)
oops
(*>*)
text {*
where @{text PUT} is a ``placeholder'' for our program under test.
*}

subsubsection {* Generating test cases *} 
text {*
Now we can automatically generate \emph{test cases}. Using the default
setup, we just apply our @{text gen_test_cases}:
*}
(*<*)
test_spec "is_sorted(PUT (l::('a list)))"
(*>*)
  apply(gen_test_cases "PUT")
txt {* 
      which leads to the test partitioning one would expect:
      @{subgoals [display, goals_limit=50]}
      Now we bind the test theorem to a particular named
       \emph{test environment}.
      *}
store_test_thm "is_sorted_result"



subsubsection {* Generating test data *}
text {* Now we want to generate concrete test data, i.e. all variables in the 
test cases must be instantiated with concrete values. This involves a 
random solver which tries to solve the constraints by randomly 
choosing values. 
*}

 gen_test_data "is_sorted_result" 
text {*
Which leads to the following test data:
@{thm [display] is_sorted_result.test_data}
*}

text {* Note that by the following statements, the test data, the test
        hypotheses and the test theorem can be inspected interactively. *}
thm is_sorted_result.test_data
thm is_sorted_result.test_hyps
thm is_sorted_result.test_thm

text {* The generated test data can be exported to an external file:*}
export_test_data "list_data.dat" is_sorted_result

subsubsection {* Mini-Example*}

fun member :: "'a \<Rightarrow> 'a list \<Rightarrow> bool" (infixl "mem" 50)
where
   "x mem S = False"
    |"x mem (y # S) = (if x = y then True else x mem S)"

thm mem_def

subsubsection {* Test Execution and Result Verification *}
text {* In principle, any SML-system should be able to run the provided
test-harness and generated test-script. Using their specific
facilities for calling foreign code, testing of non-SML programs is
possible. For example, one could test implementations written
\begin{itemize}
\item for the.Net platform,
  e.g., written in C\# using sml.net~\cite{sml.net},
\item in C using, e.g.  the foreign language
  interface of sml/NJ~\cite{smlnj} or MLton~\cite{mlton},
\item in Java using MLj~\cite{mlj}.
\end{itemize}
Depending on the SML-system, the test execution can be done within an
interpreter or using a compiled test executable.  Testing
implementations written in SML is straight-forward, based 
on automatically generated test scripts.
This generation is based on the internal code generator of
Isabelle and must be set up accordingly.
*}
consts_code "op <=" ("(_ <=/ _)")
(*<*)
 

  gen_test_script "document/list_script.sml" is_sorted_result PUT "myList.sort"  

(*>*)

text {* The key command of the generation is: *}

(*
gen_test_script "list_script.sml" is_sorted_result PUT "myList.sort"
*)

text {* which generates the following test harness:
       \lstinputlisting[style=sml]{list_script.sml} 
      *}


text {*
Further, suppose we have an ANSI C
implementation of our sorting method for sorting C arrays that we want
to test.  Using the foreign language interface provided by the SML
compiler MLton we first we have to import the sort method written in C
using the \verb|_import| keyword of MLton and further, we provide
a ``wrapper'' doing some datatype conversion, e.g.\ converting lists
to arrays and vice versa:
\begin{verbatim}
structure myList = struct
  val csort = _import "sort": int array * int -> int array;
  fun toList a = Array.foldl (op ::) [] a;
  fun sort l = toList(csort(Array.fromList(list),length l));
end
\end{verbatim}
That's all, now we can build the test executable using MLton and end
up with a test executable which can be called directly. 
\begin{table}
  \centering
{\small
\begin{verbatim}
Test Results:
=============
Test 0 -     SUCCESS, result: []
Test 1 -     SUCCESS, result: [10]
Test 2 -     SUCCESS, result: [72, 42]
Test 3 - *** FAILURE: post-condition false, result: [8, 15, -31]

Summary:
--------
Number successful tests cases:  3 of 4 (ca. 75%)
Number of warnings:             0 of 4 (ca.  0%)
Number of errors:               0 of 4 (ca.  0%)
Number of failures:             1 of 4 (ca. 25%)
Number of fatal errors:         0 of 4 (ca.  0%)

Overall result: failed 
===============
\end{verbatim}}
  \caption{A Sample Test Trace}
  \label{tab:test-trace}
\end{table}
Running our test executable will result in the test trace in
Table~\vref{tab:test-trace}. Even this small set of test vectors is sufficient to 
exploit an error in your implementation.
*}

subsection {* Improving the Testing Results *}
text {*
Obviously, in reality one would not be satisfied with the test cases generated in 
the previous section: for testing sorting algorithms one would expect that the test data
somehow represents the set of permutations of the list elements. We have already 
seen that the test specification used in the last section ``only'' enumerates
lists up to a specific length without any ordering constraints on their
elements. Thus we decide to try a more `'descriptive'' test specification that
is based on the behavior of an insertion sort algorithm:
*} 
fun   ins :: "('a::linorder) \<Rightarrow> 'a list \<Rightarrow> 'a list"
where "ins x [] = [x]"
     |"ins x (y#ys) = (if (x < y) then x#y#ys else (y#(ins x ys)))"
fun   sort:: "('a::linorder) list \<Rightarrow> 'a list"
where "sort [] = [] "
     | "sort (x#xs) = ins x (sort xs)"

text {*
Now we state our test specification by requiring that the behavior of the 
program under test @{text PUT} is identical to the behavior of our 
specified sorting algorithm @{text sort}:
*}
text {*
Based on this specification \verb|gen_test_cases| produces 
test cases representing all permutations of lists up to a fixed
length $n$. Normally, we also want to configure up to which length
lists should be generated (we call this the \emph{depth}
of test case), e.g.~we decide to generate lists up to length
$3$. Our standard setup *}

test_spec "sort l = PUT  l"

  apply(gen_test_cases "PUT")
  store_test_thm "is_sorting_algorithm0"

text {*
generates 9 test cases describing all permutations of lists of length
$1,2$ and $3$. "Permutation" means here that not only test cases (i.e.
I/O-partitions) are generated for lists of length 0, 1, 2 and 3; the
partitioning is actually finer: for two-elementary lists, for example, 
the case of a list with the first element larger or equal and the
dual case are distinguished. The entire test-theorem looks as follows:

@{thm "is_sorting_algorithm0.test_thm"}

*}


text{*A more ambitious setting is:*}


test_spec "sort l = PUT  l"

  apply(gen_test_cases 4 1 "PUT")

txt {* 
      which leads after 2 seconds to the following test partitioning (excerpt):
      @{subgoals [display, goals_limit=10]}
     *}
  store_test_thm "is_sorting_algorithm"

thm is_sorting_algorithm.test_thm

text {* In this scenario, 39 test cases are generated describing all
permutations of lists of length $1,2,3$ and $4$. "Permutation" means
here that not only test cases (i.e.  I/O-partitions) are generated for
lists of length 0, 1, 2, 3, 4; the partitioning is actually finer: for
two-elementary lists, take one case for the lists with the first
element larger or equal.  *}

text{* The case for all lists of depth 5 is feasible, however, it will
already take 8 minutes. *}


declare [[testgen_iterations=100]]
gen_test_data "is_sorting_algorithm"

thm  is_sorting_algorithm.test_data
text{* We obtain test cases like:
@{thm [display] is_sorting_algorithm.test_data}
*}

text {* If we scale down to only 10 iterations, 
this is not sufficient to solve all conditions, i.e. we 
obtain many test cases with unresolved constraints 
where @{text RSF} marks unsolved cases. In these
cases, it is unclear if the test partition is empty. Analyzing the generated
test data reveals that all cases for lists with length up to (and including) 
3 could be solved. From the 24 cases for lists of length 4 only
9 could be solved by the random solver 
(thus, overall 19 of the 34 cases 
were solved). To achieve better results, we could 
interactively increase the number of iterations which reveals that we
need to set iterations to 100 to find all solutions reliably.
\begin{center}
 \begin{tabular}{l||r|r|r|r|r|r|r|r|r}
 iterations&  5 & 10 & 20 & 25 & 30 & 40 & 50 & 75 & 100 \\
 \hline
 solved goals (of 34) & 13 & 19 & 23 & 24 & 25 & 29 & 33 & 33 &  34      
 \end{tabular}
\end{center}
Instead of increasing the number of iterations one could 
also add  other techniques such as 
        \begin{enumerate}
        \item deriving new rules that allow for 
              the generation of a simplified test theorem,
        \item introducing abstract test cases or 
        \item supporting the solving process by derived rules.
        \end{enumerate}

      *}


subsection {* Non-Inherent Higher-order Testing *}
text{*  HOL-TestGen can use test specifications that contain
higher-order operators --- although we would not claim that the test case 
generation is actually higher-order (there are no enumeration schemes
for the function space, so function variables are untreated by the
test case generation procedure so far). *}

text{* Just for fun, we reformulate the problem of finding the
maximal number in a list as a higher-order problem: *}

test_spec " foldr max l (0::int) = PUT l"
apply(gen_test_cases "PUT" simp:max_def)
store_test_thm "maximal_number"

text{* Now the test data: *}

declare [[testgen_iterations=200]]
gen_test_data "maximal_number"

thm maximal_number.test_data

end
