(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * Bank.thy --- an BirthdayBook inspired, elementary
 *              example for sequence testing based on a deterministic
 *              specification.
 *
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2009 Univ. Paris-Sud, France
 *               2009 Achim D. Brucker, Germany
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)

theory  
  Bank
imports 
  Testing
begin

declare [[testgen_profiling]]

text{* The intent of this little example is to model deposit,
check and withdraw operations of a little Bank model
in pre-postcondition style, formalize them in a setup
for \testgen{} test sequence generation and to generate elementary
test cases for it. The test scenarios will be restricted to
strict sequence checking; this excludes aspects of account
creation which will give the entire model a protocol character
(a create-operation would create an account number, and then
all later operations are just refering to this number; thus there
would be a dependence between system output and input as in
reactive sequence test scenarios.).

Moreover, in this scenario, we assume that the system under test is
deterministic.
 *}

text{* The state of our bank is just modeled by a map from
client/account information to the balance. *}
type_synonym client = string

type_synonym account_no = int

type_synonym register = "(client \<times> account_no) \<rightharpoonup> int"


subsubsection{* Operation definitions *}

text {* A standard, JML or OCL or VCC like interface specification
might look like:

\begin{verbatim}
op deposit (c : client, no : account_no, amount:nat) : unit
pre  (c,no) : dom(register)
post register'=register[(c,no) := register(c,no) + amount]

op balance (c : client, no : account_no) : int
pre  (c,no) : dom(register)
post register'=register and result = register(c,no)

op withdraw(c : client, no : account_no, amount:nat) : unit
pre  (c,no) : dom(register) and register(c,no) >= amount
post register'=register[(c,no) := register(c,no) - amount]

\end{verbatim}

*}



text{* Interface normalization turns this interface into the
following input type: *}

datatype in_c = deposit client account_no nat
              | withdraw client account_no nat
              | balance client account_no

datatype out_c = depositO| balanceO nat | withdrawO

fun    precond :: "register \<Rightarrow> in_c \<Rightarrow> bool"
where "precond \<sigma> (deposit c no m) = ((c,no) \<in> dom \<sigma>)"
    | "precond \<sigma> (balance c no) = ((c,no) \<in> dom \<sigma>)"
    | "precond \<sigma> (withdraw c no m) = ((c,no) \<in> dom \<sigma> \<and> (int m) \<le> the(\<sigma>(c,no)))"

fun    postcond :: "in_c \<Rightarrow> register \<Rightarrow> out_c \<times> register \<Rightarrow> bool"
where "postcond (deposit c no m) \<sigma> =
         (\<lambda> (n,\<sigma>'). (n = depositO \<and> \<sigma>'=\<sigma>((c,no)\<mapsto> the(\<sigma>(c,no)) + int m)))"
    | "postcond (balance c no) \<sigma> =
         (\<lambda> (n,\<sigma>'). (\<sigma>=\<sigma>' \<and> (\<exists> x. balanceO x = n \<and> x = nat(the(\<sigma>(c,no))))))"
    | "postcond (withdraw c no m) \<sigma> =
         (\<lambda> (n,\<sigma>'). (n = withdrawO \<and> \<sigma>'=\<sigma>((c,no)\<mapsto> the(\<sigma>(c,no)) - int m)))"



subsubsection{* Proving Symbolic Execution Rules for the 
               Abstractly Constructed Program *}

text{* Using the Operators \verb+impl+ and \verb+strong_impl+, we can 
synthesize an abstract program right away from the specification, i.e. the pair 
of pre- and  postcondition defined above. Since this program is even 
deterministic, we derive a set of symbolic execution rules used in the test case 
generation process which will produce symbolic results against which the PUT can 
be compared in the test driver. *}
lemma precond_postcond_implementable: 
     "implementable precond postcond"
apply(auto simp: implementable_def)
apply(case_tac "\<iota>", simp_all)
done

text{* The following lemmas reveal that this "constructed" program is actually
(due to determinism of the spec): *}

lemma impl_1:
     "strong_impl precond postcond (deposit c no m) = 
      (\<lambda>\<sigma> . if (c, no) \<in> dom \<sigma>
           then Some(depositO,\<sigma>((c, no) \<mapsto> the (\<sigma> (c, no)) + int m))
           else None)"
by(rule ext, auto simp: strong_impl_def )


lemma valid_both_spec1[simp]:
"(\<sigma> \<Turnstile> (s \<leftarrow> mbind ((deposit c no m)#S) (strong_impl precond postcond); 
                    return (P s))) = 
    (if (c, no) \<in> dom \<sigma>
     then (\<sigma>((c, no) \<mapsto> the (\<sigma> (c, no)) + int m) )\<Turnstile> (s \<leftarrow> mbind S (strong_impl precond postcond); 
                        return (P (depositO#s)))
     else (\<sigma> \<Turnstile> (return (P []))))"
by(auto simp: valid_both impl_1)


lemma impl_2:
     "strong_impl precond postcond (balance c no) = 
      (\<lambda>\<sigma>. if (c, no) \<in> dom \<sigma>
           then Some(balanceO(nat(the (\<sigma> (c, no)))),\<sigma>)
           else None)"
by(rule ext, auto simp: strong_impl_def Eps_split)

lemma valid_both_spec2 [simp]:
"(\<sigma> \<Turnstile> (s \<leftarrow> mbind ((balance c no)#S) (strong_impl precond postcond); 
                    return (P s))) = 
    (if (c, no) \<in> dom \<sigma>
     then (\<sigma> \<Turnstile> (s \<leftarrow> mbind S (strong_impl precond postcond); 
                        return (P (balanceO(nat(the (\<sigma> (c, no))))#s))))
     else (\<sigma> \<Turnstile> (return (P []))))"
by(auto simp: valid_both impl_2)


lemma impl_3:
     "strong_impl precond postcond (withdraw c no m) = 
      (\<lambda>\<sigma>. if (c, no) \<in> dom \<sigma> \<and> (int m) \<le> the(\<sigma>(c,no))
           then Some(withdrawO,\<sigma>((c, no) \<mapsto> the (\<sigma> (c, no)) - int m))
           else None)"
by(rule ext, auto simp: strong_impl_def Eps_split)

lemma valid_both_spec3[simp]:
"(\<sigma> \<Turnstile> (s \<leftarrow> mbind ((withdraw c no m)#S) (strong_impl precond postcond);
                    return (P s))) =
    (if (c, no) \<in> dom \<sigma> \<and> (int m) \<le> the(\<sigma>(c,no))
     then (\<sigma>((c, no) \<mapsto> the (\<sigma> (c, no))-int m) )\<Turnstile> (s \<leftarrow> mbind S (strong_impl precond postcond);
                        return (P (withdrawO#s)))
     else (\<sigma> \<Turnstile> (return (P []))))"
by(auto simp: valid_both impl_3)

text{* Here comes an interesting detail revealing the power of the
approach: The generated sequences still respect the preconditions
imposed by the specification  - in this case, where we are talking about
a client for which a defined account exists and for which we will never
produce traces in which we withdraw more money than available on it. *}

subsubsection{* Test Specifications *}


fun test_purpose :: "[client, account_no, in_c list] \<Rightarrow> bool"
where
  "test_purpose c no [] = False"
| "test_purpose c no (a#R) = (case R of 
                               [] \<Rightarrow> a = balance c no
                              | a'#R' \<Rightarrow> (((\<exists> m. a = deposit c no m) \<or> 
                                          (\<exists> m. a = withdraw c no m)) \<and>
                                          test_purpose c no R))"

lemma inst_eq : "f(x,y) = Some z \<Longrightarrow> f(x,y) = Some z"
by auto
lemma inst_eq2 : "x = y \<Longrightarrow> x = y"
by auto



test_spec test_balance:
assumes account_defined: "(c,no) \<in> dom \<sigma>_0" 
(* and     store_finite   : "\<sigma>_0 = map_of T" *)
and     test_purpose   : "test_purpose c no S"
and     symbolic_run_yields_x :
                         "\<sigma>_0 \<Turnstile> (s \<leftarrow> mbind S (strong_impl precond postcond);
                                         return (s = x))"
shows  " \<sigma>_0 \<Turnstile> (s \<leftarrow> mbind S PUT; return (s = x))"
apply(insert (* store_finite *)  account_defined test_purpose symbolic_run_yields_x)
apply(gen_test_cases "PUT" split: HOL.split_if_asm)
(* some massage by hand ... should be automated at some point ...*)
(*FIXME broken
apply(drule_tac f="empty((?X1, ?X2) \<mapsto> 2)" and x="?X1" and y="?X2" and z="2" in inst_eq)
apply (simp)  
(* Ich bin ehrlich gesagt nicht ganz mit dem Resultat zufrieden. 
   Zu erwarten waere: X[(c,no) \<mapsto> k] = sigma, statt dessen haben wir
   was viel spezielleres. Aber funktionieren wird es damit . *)
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 5]" and x="?X1" and y="?X2" and z="5" in inst_eq, simp) 
(* seems to be oversimplified, but ok *)
apply(drule_tac x="?X1" and y="?X1" in inst_eq2, simp)
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 3]" and x="?X1" and y="?X2" and z="3" in inst_eq, simp) 
defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 7]" and x="?X1" and y="?X2" and z="7" in inst_eq, simp) 
apply(drule_tac x="?X1" and y="?X1" in inst_eq2, simp)
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 6]" and x="?X1" and y="?X2" and z="6" in inst_eq, simp) 
defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 9]" and x="?X1" and y="?X2" and z="9" in inst_eq, simp)
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 10]" and x="?X1" and y="?X2" and z="10" in inst_eq, simp) 
apply(drule_tac x="?X1" and y="?X1" in inst_eq2, simp)
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 4]" and x="?X1" and y="?X2" and z="4" in inst_eq, simp) 
defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 5]" and x="?X1" and y="?X2" and z="5" in inst_eq, simp) 
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 2]" and x="?X1" and y="?X2" and z="2" in inst_eq, simp) 
apply(drule_tac x="?X1" and y="?X1" in inst_eq2, simp)
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 2]" and x="?X1" and y="?X2" and z="2" in inst_eq, simp)
defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 5]" and x="?X1" and y="?X2" and z="5" in inst_eq, simp)
apply(drule_tac x="?X1" and y="?X1" in inst_eq2, simp)
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 2]" and x="?X1" and y="?X2" and z="2" in inst_eq, simp)
defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 3]" and x="?X1" and y="?X2" and z="3" in inst_eq, simp)
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 3]" and x="?X1" and y="?X2" and z="3" in inst_eq, simp)
apply(drule_tac x="?X1" and y="?X1" in inst_eq2, simp)
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 3]" and x="?X1" and y="?X2" and z="3" in inst_eq, simp)
apply(drule_tac x="?X1" and y="?X1" in inst_eq2, simp)
defer 1 defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 3]" and x="?X1" and y="?X2" and z="3" in inst_eq, simp)
defer 1
apply(drule_tac f="[(?X1, ?X2) \<mapsto> 6]" and x="?X1" and y="?X2" and z="6" in inst_eq, simp)
defer 1 defer 1
defer 1
*)
store_test_thm "bank"

text{* This results in a test-space exploration:
\begin{verbatim}
 1. (\<lambda>a. Some 2) \<Turnstile>
    ( s \<leftarrow> mbind [balance ?X2X1233 ?X1X1231] PUT; return s = [balanceO 2])
 2. THYP
     ((\<exists>x xa xb xc.
          xb (xa, x) = Some xc \<longrightarrow>
          (xb \<Turnstile>
           ( s \<leftarrow> mbind [balance xa x]
                  PUT; return s = [balanceO (nat xc)]))) \<longrightarrow>
      (\<forall>x xa xb xc.
          xb (xa, x) = Some xc \<longrightarrow>
          (xb \<Turnstile>
           ( s \<leftarrow> mbind [balance xa x]
                  PUT; return s = [balanceO (nat xc)]))))
 3. (\<lambda>a. Some 5) \<Turnstile>
    ( s \<leftarrow> mbind
           [deposit ?X2X1190 ?X1X1188 ?X5X1196, balance ?X2X1190 ?X1X1188]
           PUT; return s = [depositO, balanceO (nat (5 + int ?X5X1196))])
 4. THYP
     ((\<exists>x xa xb xc xd xe.
          xb (xa, x) = Some xc \<longrightarrow>
          xb (xa, x) = Some xe \<longrightarrow>
          (xb \<Turnstile>
           ( s \<leftarrow> mbind [deposit xa x xd, balance xa x]
                  PUT; return s =
                              [depositO, balanceO (nat (xe + int xd))]))) \<longrightarrow>
      (\<forall>x xa xb xc xd xe.
          xb (xa, x) = Some xc \<longrightarrow>
          xb (xa, x) = Some xe \<longrightarrow>
          (xb \<Turnstile>
           ( s \<leftarrow> mbind [deposit xa x xd, balance xa x]
                  PUT; return s =
                              [depositO, balanceO (nat (xe + int xd))]))))
 5. THYP
     ((\<exists>x xa xb xc xd.
          \<not> (\<exists>y. xb (xa, x) = Some y) \<longrightarrow>
          xb (xa, x) = Some xd \<longrightarrow>
          (xb \<Turnstile>
           ( s \<leftarrow> mbind [deposit xa x xc, balance xa x]
                  PUT; return s = []))) \<longrightarrow>
      (\<forall>x xa xb xc xd.
          \<not> (\<exists>y. xb (xa, x) = Some y) \<longrightarrow>
          xb (xa, x) = Some xd \<longrightarrow>
          (xb \<Turnstile>
           ( s \<leftarrow> mbind [deposit xa x xc, balance xa x]
                  PUT; return s = []))))
 6. int ?X5X1088 \<le> 7 \<Longrightarrow>
    (\<lambda>a. Some 7) \<Turnstile>
    ( s \<leftarrow> mbind
           [withdraw ?X2X1082 ?X1X1080 ?X5X1088, balance ?X2X1082 ?X1X1080]
           PUT; return s = [withdrawO, balanceO (nat (7 - int ?X5X1088))])
 7. THYP
     ((\<exists>x xa xb xc xd xe.
          xb (xa, x) = Some xc \<longrightarrow>
          int xd \<le> xe \<longrightarrow>
          xb (xa, x) = Some xe \<longrightarrow>
          (xb \<Turnstile>
           ( s \<leftarrow> mbind [withdraw xa x xd, balance xa x]
                  PUT; return s =
                              [withdrawO,
                               balanceO (nat (xe - int xd))]))) \<longrightarrow>
      (\<forall>x xa xb xc xd xe.
          xb (xa, x) = Some xc \<longrightarrow>
          int xd \<le> xe \<longrightarrow>
          xb (xa, x) = Some xe \<longrightarrow>
          (xb \<Turnstile>
           ( s \<leftarrow> mbind [withdraw xa x xd, balance xa x]
                  PUT; return s =
                              [withdrawO, balanceO (nat (xe - int xd))]))))
 8. THYP
     ((\<exists>x xa xb xc xd.
          \<not> (\<exists>y. xb (xa, x) = Some y) \<longrightarrow>
          xb (xa, x) = Some xd \<longrightarrow>
          (xb \<Turnstile>
           ( s \<leftarrow> mbind [withdraw xa x xc, balance xa x]
                  PUT; return s = []))) \<longrightarrow>
      (\<forall>x xa xb xc xd.
          \<not> (\<exists>y. xb (xa, x) = Some y) \<longrightarrow>
          xb (xa, x) = Some xd \<longrightarrow>
          (xb \<Turnstile>
           ( s \<leftarrow> mbind [withdraw xa x xc, balance xa x]
                  PUT; return s = []))))
 9. \<not> int ?X4X980 \<le> 9 \<Longrightarrow>
    (\<lambda>a. Some 9) \<Turnstile>
    ( s \<leftarrow> mbind [withdraw ?X2X976 ?X1X974 ?X4X980, balance ?X2X976 ?X1X974]
           PUT; return s = [])
 10. THYP
      ((\<exists>x xa xb xc xd.
           \<not> int xc \<le> xd \<longrightarrow>
           xb (xa, x) = Some xd \<longrightarrow>
           (xb \<Turnstile>
            ( s \<leftarrow> mbind [withdraw xa x xc, balance xa x]
                   PUT; return s = []))) \<longrightarrow>
       (\<forall>x xa xb xc xd.
           \<not> int xc \<le> xd \<longrightarrow>
           xb (xa, x) = Some xd \<longrightarrow>
           (xb \<Turnstile>
            ( s \<leftarrow> mbind [withdraw xa x xc, balance xa x]
                   PUT; return s = []))))

\end{verbatim}
*}

declare [[testgen_iterations=50]]
gen_test_data "bank"

thm bank.test_data

end