(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * Triangle_test.thy --- testing the classical triangle example
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2003-2007 ETH Zurich, Switzerland
 *               2009-2010 Achim D. Brucker, Germany
 *               2009-2010 Univ. Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: Triangle_test.thy 9041 2010-11-10 17:40:39Z krieger $ *)


header {* Testing the Triangle Example *}

theory 
  Triangle_test
imports
  Triangle
  Testing
begin

declare [[testgen_profiling]]

text{* The test theory \texttt{Triangle\_test} is used to demonstrate 
       the pragmatics  of \testgen{} in the standard triangle example;
       The demonstration elaborates three test plans: standard test 
       generation (including test driver generation), abstract test data
       based test generation, and abstract test data based
       test generation reusing partially synthesized abstract
       test data. *}


section {* The Standard Workflow *}

    text {* We start with stating a test specification for
       a program under test: it must behave as specified in the definition 
       of @{text classify_triangle}.

       Note that the variable @{term program} is used to label an
       arbitrary implementation of the current \index{program under
       test}program under test that should fulfill the test
       specification: *}


test_spec "program(x,y,z) = classify_triangle x y z"

    txt {*
    By applying @{term gen_test_cases} we bring the proof state into 
    testing normal form (TNF). 
    *}

  apply(gen_test_cases  "program" simp add: triangle_def
                                            classify_triangle_def)

    (* 35 Goals: 17 THYP / 18 test cases *)
    txt {* In this example, we decided to generate symbolic test cases
    and to unfold the triangle predicate 
    by its definition before the process. This leads to a formula with, 
    among others, the  following clauses: 
    @{subgoals [display, goals_limit=5]} 
    *}
    (*\
    Exercise: compare  these with the  results of Dick and 
    Faivre~\cite{dick.ea:testing:1993}.
    *)
    txt {* Note that the
    computed \TNF is not minimal, i.e. further simplification and
    rewriting steps are needed to compute the \emph{minimal set of
    symbolic test cases}. The following post-generation simplification
    improves the generated result before ``frozen'' into a 
    \emph{test theorem}: 
    *}

  apply(simp_all) 

    txt {*
    Now, ``freezing'' a test theorem\index{test theorem} technically means 
    storing it
    into a specific data structure provided by \testgen, namely a \emph{test
    environment}\index{test environment} that captures all data relevant to a
    test: *}

  store_test_thm "triangle_test"


text {* The resulting test theorem is now bound to a particular name
       in the Isar environment, such that it can inspected by the
       usual Isar command \verb+thm+. *}
 
thm "triangle_test.test_thm"

text{*  We compute the concrete \emph{test statements} by instantiating
        variables by constant terms in the symbolic test cases for
        ``@{term program}'' via a random test procedure:
     *}

gen_test_data "triangle_test"


text {* which results in  @{thm[display] triangle_test.test_data } *}

thm "triangle_test.test_hyps" 
thm "triangle_test.test_data"

text{*  Now we use the generated test data statement lists to automatically
        generate a test driver, which is controlled by the test harness.
        The first argument is the external SML-file name into which the 
        test driver is generated, the second argument the name of the
        test data statement set and the third the name of the (external)
        program under test:
     *}

gen_test_script "triangle_script.sml" "triangle_test" "program"


section {* The Modified Workflow: Using Abstract Test Data *}

text {* 
  There is a viable alternative to the standard development process above: 
  instead of unfolding triangle and trying to generate ground substitutions
  satisfying the constraints, one may keep triangle in the test theorem, treating
  it as a building block for new constraints. 
  Such building blocks will also be called 
  \index{abstract test case}\emph{abstract test cases}.

  In the following, we will set up a new version of the
  test specification, called @{term triangle2}, and prove the
  relevant abstract test cases individually before test case 
  generation. These proofs are highly automatic, but the choice 
  of the abstract test data in itself is ingenious,
  of course. 
*}
(*
  Nevertheless, the computation for establishing if
  a certain triple is encapsulated in these proofs, deliberating the 
  main test case generation of @{term triangle2} from them. 
  In fact, these contain 5 arithmetic constraints which represent 
  already a sensible load if given to the random solver. 
*)
text {*
  The abstract test data will be assigned to the subsequent
  test generation for the test specification @{term triangle2}.
  Then the test data generation phase is started for @{term triangle2}
  implicitly using the abstract test cases. The association 
  established by this assignment is also stored in the test
  environment.

  The point of having abstract test data is that it can be generated 
  ``once and for all'' and inserted before the test data selection 
  phase producing a ``partial'' grounding. It will turn out that the 
  main state explosion is shifted from the test case generation to the 
  test data selection phase. 
  *}

subsection{* The ``ingenious approach'' *}

lemma triangle_abscase1 [test "triangle2"]: "triangle 1 1 1"
  by(auto simp: triangle_def)

lemma triangle_abscase2 [test"triangle2"]:"triangle 1 2 2"
  by(auto simp: triangle_def)

lemma triangle_abscase3 [test"triangle2"]:"triangle 2 1 2"
  by(auto simp: triangle_def)

lemma triangle_abscase4 [test"triangle2"]:"triangle 2 2 1"
  by(auto simp: triangle_def)

lemma triangle_abscase5 [test"triangle2"]:"triangle 3 4 5"
  by(auto simp: triangle_def)

lemma triangle_abscase6 [test"triangle2"]:"\<not> triangle -1 1 2"
  by(auto simp: triangle_def)

lemma triangle_abscase7 [test"triangle2"]:"\<not> triangle 1 -1 2"
  by(auto simp: triangle_def)

lemma triangle_abscase8 [test"triangle2"]:"\<not> triangle 1 2 -1"
  by(auto simp: triangle_def)

lemma triangle_abscase9 [test "triangle2"]: "\<not> triangle -1 -1 -1"
  by(auto simp: triangle_def)

lemma triangle_abscase10 [test "triangle2"]: "\<not> triangle -1 1 -1"
  by(auto simp: triangle_def)

lemma triangle_abscase11 [test "triangle2"]: "\<not> triangle 1 -1 -1"
  by(auto simp: triangle_def)

lemma triangle_abscase12 [test "triangle2"]: "\<not> triangle -1 -1 1"
  by(auto simp: triangle_def)
   


lemmas abs_cases = triangle_abscase1 triangle_abscase2 triangle_abscase3 triangle_abscase4
                   triangle_abscase5 triangle_abscase6 triangle_abscase7 triangle_abscase8
                   triangle_abscase9 triangle_abscase10 triangle_abscase11 triangle_abscase12 

text{* Just for demonstration purposes, we apply the abstract test data solver
directly in the proof: *}

test_spec "prog(x,y,z) = classify_triangle x y z"
  apply(gen_test_cases "prog" simp add: classify_triangle_def)
  apply(tactic "ALLGOALS(TestGen.ONLY_POS (TestGen.TRY' (EqSubst.eqsubst_tac @{context} [0] [@{thm PO_def}] 
    THEN' TestGen.abs_data_tac @{context} @{thms abs_cases})))")
oops 


test_spec "prog(x,y,z) = classify_triangle x y z"
  apply(gen_test_cases "prog" simp add: classify_triangle_def)
  store_test_thm "triangle2"
thm "triangle2.test_thm"


gen_test_data "triangle2"

text{* The test data generation is started and implicitly uses 
       the abstract test data assigned to the test theorem 
       \emph{triangle2}. Again, we inspect the results:
@{thm[display] triangle2.test_data} *}

 
thm "triangle2.test_hyps" 
thm "triangle2.test_data"

subsection{* Alternative: Synthesizing Abstract Test Data *}

text {* In fact, part of the ingenious work of generating abstract
        test data can be synthesized by using the test case generator
        itself. This usage scenario proceeds as follows:
        \begin{enumerate}
        \item we set up a decomposition of @{term triangle}
              in an equality to itself; this identity is disguised
              by introducing a variable @{term prog} which is stated 
              equivalent to @{term triangle} in an assumption,  
        \item the introduction of this assumption is delayed; i.e.
              the test case generation is performed in a state
              where this assumption is not visible,
        \item after executing test case generation, we fold back 
              @{term prog} against @{term triangle}.
        \end{enumerate}
  *}

test_spec abs_triangle : 
assumes 1: "prog = triangle" 
shows      "triangle x y z = prog x y z"
  apply(gen_test_cases "prog" simp add: triangle_def)
  apply(simp_all add: 1)
store_test_thm "abs_triangle"

thm abs_triangle.test_thm  

text {* which results in  @{thm[display] abs_triangle.test_thm} Thus, we
        constructed  test cases for being triangle or not in 
        terms of arithmetic constraints. These are amenable to
        test data generation by increased random solving, which
        is controlled by the test environment variable \verb+iterations+:*}

declare [[testgen_iterations=100]]
gen_test_data "abs_triangle"
text {* resulting in: @{thm[display] abs_triangle.test_data} *}


text {* Thus, we achieve solved ground instances for abstract test data. 
        Now, we assign these synthesized test data to the new future
        test data generation. Additionally to the synthesized abstract
        test data, we assign the data for isosceles and equilateral
        triangles; these can not be revealed from our synthesis since
        it is based on a subset of the constraints available in the
        global test case generation. *}


declare abs_triangle.test_data[test"triangle3"]
declare triangle_abscase1[test"triangle3"]
declare triangle_abscase2[test"triangle3"]
declare triangle_abscase3[test"triangle3"]


text {* The setup of the testspec is identical as for triangle2; it is 
        essentially a renaming. *}
     
test_spec "program(x,y,z) = classify_triangle x y z"
  apply(simp add: classify_triangle_def)
  apply(gen_test_cases  "program" simp add: classify_triangle_def)
  store_test_thm "triangle3"


text{* The test data generation is started again on the
       basis on synthesized and selected hand-proven abstract 
       data. *}

declare [[testgen_iterations=10]]
gen_test_data "triangle3"

thm "triangle3.test_hyps" 
thm "triangle3.test_data"
(*
text {* resulting in: @{thm[display] triangle3.test_data} *}
*)
end
