(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * TestGen.ML  --- test data/case generation routines
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2003-2010 ETH Zurich, Switzerland
 *               2003-2010 Achim D. Brucker, Germany
 *               2003 2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: TestGen.ML 9332 2012-03-26 15:17:44Z krieger $ *)

(* Module Test *)

infix 1 |$>

infix 6 DOWNTO

infix 0 COLLECT_SS

signature TESTGEN =
sig
   val  mt_testenv        : TestEnv.testenv

   val trace              : bool Config.T
   (* switches on debug information *)

   val profiling          : bool Config.T
   (* switches on timer profiler *)

   val no_finalize        : bool Config.T
   (* don't generate constraints (POs) *)

   val is_test : term -> bool

   val |$> : thm * tactic -> thm
   val TRY' : (int -> tactic) -> int -> tactic
   val COND' : (int -> thm -> bool) -> (int -> tactic) -> (int -> tactic) -> int -> tactic
   val DOWNTO : int * int -> (int -> tactic) -> tactic
   val DEPTH_SPLIT : (int -> tactic) -> int -> tactic

   val ALLCASES : (int -> tactic) -> tactic

   val ONLY_POS : (int -> tactic) -> int -> tactic

   val finalize_tac : Proof.context -> string list -> tactic

   val abs_data_tac : Proof.context -> thm list -> int -> tactic

   val mp_fy : int -> tactic
   val all_ify : Proof.context -> string list -> int -> tactic
   val allify_frees_tac : Proof.context -> string list -> int -> tactic
   val thyp_ify : Proof.context -> int -> tactic

   val var_hyp_subst_tac : int -> tactic

   val gen_test_case_tac : Proof.context -> string list -> tactic
   (* "gen_test_case_tac ctxt prognames" is a
      tactic that converts all subgoals of current proof state into
      test-normal form - i.e. a conjunctive normal form
      (instead of DNF), where the conclusions have the
      form "P(progname x1 .. xn)", where progname does
      not occur in the premisses (the "constraints")
      of the test. As a prerequisite, a data-separation
      theorem of depth k is resolved for up to >breadth< number
      of free variables in the current proof-state.

      NOTE: At present, the generation of data-separation
      lemmas is limited to "real" free variables and will
      not work for parameters of a test specification. Further,
      the proof-state is assumed to have only one goal.
    *)

   (* Included for complete backward compatibility *)
   val gen_test_case_old_tac : Proof.context -> string list  -> tactic

   val gen_test_data  :  string -> Context.generic -> Context.generic
   (* "gen_test_data name context" constructs ground type instances of
      test theorems associated to "name". In a further step,
      "abstract" or local test data were resolved into the resulting test
      theorems (if available in the testenv under "name"), remaining
      meta variables were instantiated by a "random_solver"
      that attempts to generate an instance and tries to solve
      the constraints of a goal up to "iterations" times.

      The resulting simplified and "solved" test data theorems
      were stored again in the test environment.
    *)

   val get_test_data      : theory -> string  -> thm list
   (* get_test_data thy name extracts the test judgements
      (i.e. the premisses of test data statement, that are not
      test hypothesis or proof obligation.). *)

   val get_test_hyps      : theory -> string -> thm list
   (* get_test_hyps thy name extracts the test hypothesis
      of a !!test data statement!!, not a test theorem. *)

   val get_pos      : theory -> string -> thm list
   (* extract remaining constraints (proof obligations) *)

   val animate            : theory -> string  -> unit
   (* ground test data statements may be animated, if "P(progname x1 .. xn)"
      is executable. "animate" presents test data statements by
      unfolding "progname x1 .. xn" and simplifying on a stepwise
      basis. (* for the moment, it's just asm_simp_tac with the
      standard evaluation trace, in the future its more or less
      the same except with slightly improved output *)
    *)

    val gen_testscript: string -> string -> string -> string -> theory -> theory


    val setup: theory -> theory
end

structure TestGen : TESTGEN =
struct

open HOLogic;
open TestEnv;
open Clocks;

val (trace, trace_setup) = Attrib.config_bool @{binding testgen_trace} (K false);
val (profiling, profiling_setup) = Attrib.config_bool @{binding testgen_profiling} (K false);
val (no_finalize, no_finalize_setup) = Attrib.config_bool @{binding testgen_no_finalize} (K false);


(* ************************************************************************* *)
(* access to datatype package ...                                            *)
(* ************************************************************************* *)

fun isDataType thy (Type(tn,_)) = is_some (Datatype.get_info thy tn)
   |isDataType thy _ = false;


fun get_case_split_of_dt thy type_name = #exhaust (Datatype.the_info thy type_name);

fun get_nchotomy_of thy type_name = #nchotomy (Datatype.the_info thy type_name);


(* ************************************************************************* *)
(* access to record package ...                                              *)
(* ************************************************************************* *)

fun isRecordType thy (Type(tn,_)) = is_some (Record.get_info thy tn)
   |isRecordType thy _ = false;
   
fun get_case_split_of_record thy type_name = #cases(the (Record.get_info thy type_name));

(* ************************************************************************* *)
(* term computations ...                                                     *)
(* ************************************************************************* *)

fun max_depth (Free h) = (0,[h])
   |max_depth (Const h)= (0,[])
   |max_depth (Abs(s,tt,t)) = max_depth t
   |max_depth (s $ t) = let val (k,S) = max_depth (s)
                            val (k',S') = max_depth (t)
                        in  if k > k'+1 then (k,S)
                            else if k = k'+1 then (k, S @ S')
                                 else (k'+1, S')
                        end
   |max_depth (_) = (0,[]);

(* relies on definitions in Testing.thy ... *)
val is_thyp = exists_Const (fn (@{const_name THYP},_) => true | _ => false)

fun is_po (Const (@{const_name Trueprop},_) $ (Const (@{const_name PO},_) $ _)) = true
   |is_po _   = false;

fun is_test t = (not (is_thyp t)) andalso (not (is_po t))

fun exists_prog prog = Term.exists_subterm
                           (fn Free (x, _) => member (op =) prog x
                             | Const (c, _) => member (op =) prog (Long_Name.base_name c) (* FIXME proper internal name!? *)
                             | _ => false);

fun calc_constr_list tgt descr =
let fun DtTFree_tvar x = (Datatype_Aux.dest_DtTFree x, HOLogic.typeS)
    val sorts          = maps (fn (_,(_,s,_)) => map DtTFree_tvar s) descr
    val recTs          = Datatype_Aux.get_rec_types descr sorts;
    val newTs          = Library.take (length descr) recTs;
    val (_,(_,insts,tgt_constrs)) = hd(filter (fn (_,(n,_,_)) => n = tgt) descr)
    val T              = hd(filter (fn (Type(n,_)) => n = tgt) newTs)
    val typ_of_dtyp    = Datatype_Aux.typ_of_dtyp descr sorts
    val constr_decls   = map (fn (cname, cargs) =>
                                 Const(cname, map typ_of_dtyp cargs ---> T))
                             (tgt_constrs)
in  (map Datatype_Aux.dest_DtTFree insts, constr_decls) end;



fun random_term' w n max thy cod_term_tab (Type(s,S)) =
(* w => the weight on the actual level, initial value 1
   n => level counter, inital value 0
   max => maximal allowed number of levels
*)
       (case Symtab.lookup cod_term_tab s of
         NONE   => (* default random term generator ...
                      TODO : should do also something for functions ... *)
                   (case Type(s,S) of
                      Type(@{type_name int},_) => mk_number intT (IntInf.fromInt((random_range 0 20) - 10))
                    (* FIXME redundant!? -- since Isabelle2008 'a set merely abbreviates 'a => bool *)
                    (* FIXME after Isabelle2011-1 'a set will be again a proper type constructor*)
                    | Type("set",_) => Const(@{const_name set},dummyT) $
                                    (random_term' w n max thy cod_term_tab (Type(@{type_name list},S)))
                    | Type(@{type_name fun},[T, Type(@{type_name bool}, [])]) =>
                      Const(@{const_name set}, Type(@{type_name fun}, [Type(@{type_name list}, [T]), Type(s,S)])) $
                           (random_term' w n max thy cod_term_tab (Type(@{type_name list},[T])))
                    | Type(@{type_name fun},[T,U]) => absdummy T (random_term' w n max thy cod_term_tab U)
                    | _     => let
                                   val descr = #descr(Datatype.the_info thy s)

                                   val (insts,constrs) = calc_constr_list s descr

                                   val weighed_constrs = let
                                       fun ar args = (length
                                                              (filter (fn t =>
                                                                          case t of
                                                                              Datatype_Aux.DtRec _ => true
                                                                            | _ => false ) args) )
                                       val constr_arity_list =  map (fn (f,args) => (f,(ar args)))
                                                                  (maps (#3 o snd) descr)
                                   in
                                       map (fn (f,a) => if a = 0 then (1,f) else (a * w,f))
                                           constr_arity_list
                                   end

                                   val  weighed_constrs = if (n >= max)
                                                          then filter (fn (w,f) => w =1) weighed_constrs
                                                          else weighed_constrs
                                   fun weight_of t = fst(hd ((filter (fn (w,ty) => ty=t)) weighed_constrs))

                                   fun frequency' xs = let
                                       val max = List.foldl Int.max 0 (map fst xs);
                                       val xs' = map (fn (x,a) => (max-x+1,a)) xs
                                   in
                                       frequency xs'
                                   end
                                   (* the default is a random bias towards constants *)
                                   val constr = frequency' weighed_constrs
                                   val Const(h,ty) = hd (filter (fn Const(h,ty) => h = constr) constrs)
                                   val w          = weight_of h
                                   val ty_binds   = insts ~~ S
                                   fun ty_inst (s,_) = the (AList.lookup (op =) ty_binds s)
                                   val instantiated_ty = map_type_tfree ty_inst ty
                                   val const_head = Const(h,instantiated_ty)
                                   val arg_ty = binder_types instantiated_ty
                               in  list_comb(const_head,
                                             (map(random_term' w (n+1) max thy cod_term_tab)(arg_ty)))
                               end (* FIXME avoid handle _ *)
                               handle _=>error("Cannot generate random value for type:" ^s))
       | SOME R => R S)
   |random_term' _ _ _ _ _ _ = error "Internal error in random_term: type not ground";


fun random_term thy cod_term_tab typ = random_term' 1 0 10000 thy cod_term_tab typ

(* test section:

val ttt = [HOLogic.intT,HOLogic.unitT,HOLogic.boolT,
           HOLogic.mk_setT HOLogic.intT,
           HOLogic.listT HOLogic.intT];
map (random_term Symtab.empty) ttt;

 *)


(* complex type instantiations ... *)

fun groundT_thm thy bound candTs thm =
    let val thm   = Thm.varifyT_global thm
        val tvars = Term.add_tvars (list_comb(concl_of thm,prems_of thm)) []
        fun ctyp_of_sort x so t = if Sign.of_sort thy (t,so)
                                  then (SOME(ctyp_of thy (TVar(x,so)), ctyp_of thy t)
                                             handle _ => NONE)  (* FIXME avoid handle _ *)
                                  else NONE
        fun prodS [] B = []
           |prodS (a::R) B =  (map(fn x => a::x) B) @ (prodS R B)

        fun compute_instsT [] = [[]]
           |compute_instsT ((x,so)::R) = prodS (List.mapPartial (ctyp_of_sort x so) candTs)
                                               (compute_instsT R)
    in
        map (fn x => Drule.instantiate_normalize (x,[]) thm)
            (Library.take bound (compute_instsT tvars))
    end;

(* test

 val thm' = groundT_thm thy
                        (#type_range_bound(rep_testenv te))
                        (#type_candicates(rep_testenv te))
                        thm

 *)

(* *********************************************************************** *)
(*                                                                         *)
(* SOME more universal tacticals -> tctical.ML                             *)
(*                                                                         *)
(* *********************************************************************** *)

fun thm |$> tac =
    case Seq.pull (tac thm) of
         SOME (m,_) => m
        |_ => error"|$> : TACTIC FAILED"

fun TRY' tac n = TRY(tac n) (* standard combinator,
                               not implemented so far in Isabelle *)


fun COND' testfun thenf elsef x prf =
      if testfun x prf then thenf x prf else elsef x prf;


fun (k DOWNTO (n:int)) = fn tac => (fn st =>
  let fun doall x = if x < n
                    then all_tac
                    else ((tac x) THEN doall(x-1))
      val m = nprems_of st
  in  if k <= 0 orelse k > m
      then all_tac st
      else doall k st
  end)


fun DEPTH_SPLIT tac n st =
(* PRE tac should only exhausting - i.e. increase
       the number of subgoals finitely many times
   POST if tac never fails, DEPTH_SPLIT tac will not fail.
 *)
    let val m   = nprems_of st
        val res = st |$> (tac n)
        val m'  = nprems_of res
    in  if m' > m
        then ((n + m' - m) DOWNTO n)(DEPTH_SPLIT tac) (res)
        else all_tac res
    end

fun SOLVE_ASMS ctxt tac n state =
    let val goal = Logic.get_goal (prop_of state) n
        val params = Logic.strip_params goal
        val prems =  map HOLogic.dest_Trueprop (rev(Logic.strip_imp_prems goal))
        val eqTrueI2 = @{lemma "[|P;P |] ==> True" by auto}
        fun decomp thm n = let val mm = (thm RS conjunct1)
                           in  (dtac (mm RS eqTrueI2) n) THEN
                               (etac @{thm TrueE} n) THEN
                               (decomp (thm RS conjunct2) n) end
                           handle THM _ => (dtac (thm RS eqTrueI2) n) THEN
                                           (etac @{thm TrueE} n)

    in  case prems of
        [] => all_tac state
       |a::r => let val prem = List.foldl HOLogic.mk_conj a r
                   val prem'= (Term.subst_bounds ((rev(map Free params)),prem))
                   val thm  = Goal.prove ctxt [] [] (HOLogic.mk_Trueprop prem') tac
                in (decomp thm n) state end
                handle _ => no_tac state (* could not solve asms ... *)  (* FIXME avoid handle _ *)
    end


fun auto_solver thms {prems: thm list, context: Proof.context} state =
    let val thm'  = state |$> (TRY(safe_tac (context delrules [notI])))
        val goals = Thm.nprems_of thm'
        fun solve_thm thm = FIRSTGOAL (fn n => (rtac thm n) THEN
                                       (IF_UNSOLVED (SOLVE(auto_tac context))))
        fun solve_thms [] = no_tac
           |solve_thms (a::R) = solve_thm a ORELSE solve_thms R
    in  if goals > 0 then (solve_thms thms) thm' else all_tac thm'
    end


(* *********************************************************************** *)
(*                                                                         *)
(* SOME Testing specific tacticals                                         *)
(*                                                                         *)
(* *********************************************************************** *)

(* PRE n>0 and n <= nprems_of thm *)
fun is_THYP n thm = is_thyp(Logic.nth_prem(n, prop_of thm))

fun is_PO n thm = is_po(Logic.nth_prem(n, prop_of thm))

fun is_TEST n thm = is_test(Logic.nth_prem(n, prop_of thm))

fun ONLY_CASES tac = COND' is_TEST tac (K all_tac)

fun ALLCASES_before k tac = (k DOWNTO 1) (ONLY_CASES tac)

fun ALLCASES tac thm = (ALLCASES_before (nprems_of thm) tac thm)

fun ONLY_POS tac = COND' is_PO tac (K all_tac)


fun unfold rules = asm_full_simp_tac (HOL_basic_ss addsimps rules)

(* Optimization potential: safe_tac against something more
   specialized should do ... *)
fun if_splitter cs  = asm_full_simp_tac(HOL_basic_ss addsplits if_splits)
                   THEN' TRY' (fn n => SELECT_GOAL (safe_tac cs) n)



(* *********************************************************************** *)
(*                                                                         *)
(* basic tactics ...                                                       *)
(*                                                                         *)
(* *********************************************************************** *)

fun mp_fy n = (REPEAT(etac rev_mp n));

fun all_ify ctxt excpt n state =
    let val subgoal   = Logic.nth_prem(n, prop_of state)
        val params    = Logic.strip_params subgoal
        val frees     = Term.add_frees subgoal []
        fun sbs(n,t)  = res_inst_tac ctxt [(("x",0), n)] spec
        val vars      = filter_out (member (op =) excpt o fst) (params @ frees)
    in  (if null vars then no_tac
         else List.foldr(fn(x,ta) => (sbs x n) THEN ta)
                   prune_params_tac vars)
        (state)
    end

(*FIXME Free's are actually local consts, and cannot be closed in general
  without disrupting Isar locality*)
fun allify_frees_tac ctxt excpt = SUBGOAL (fn (subgoal, i) =>
    let val thy = ProofContext.theory_of ctxt
        val a = (("'a", 0), @{sort type})
        fun spec_tac (x, T) =
            rtac (Thm.certify_instantiate ([(a, T)], [((("x", 0), T), Free (x, T))])
                    (Thm.transfer thy @{thm spec})) i
        val vars = filter_out (member (op =) excpt o fst) (Term.add_frees subgoal [])
    in
        if null vars then all_tac
        else EVERY (map spec_tac vars)
    end)

fun thyp_ify ctxt n =
    EVERY[mp_fy n, all_ify ctxt [] n, rtac (@{thm THYP_def} RS @{thm iffD1}) n]

fun vars_to_frees t = let
    val termvars = Term.add_vars t []
    val freenames = Term.add_free_names t []
    val varnames = map (fn ((s, a),_) => s^(Int.toString a)) termvars
    val newfreenames = Name.variant_list freenames varnames
    val insts = map (fn (x as (_, T), freename) => (Var x, Free (freename, T))) (termvars ~~ newfreenames)
in
    Term.subst_atomic insts t
end

fun convert_goals_to_metahyps (test:(term -> bool)) thm =
    let val thy = theory_of_thm thm
    in  map ((fn prop => Thm.forall_elim_vars 0 (Thm.assume (cterm_of thy prop))) o vars_to_frees)
            (filter test (prems_of thm))
    end;


fun instantiate_tac insts n thm =
    let val goal = cterm_of (theory_of_thm thm) (Logic.nth_prem(n, prop_of thm))
    in  (rtac(Drule.instantiate_normalize ([],insts)(Thm.trivial goal)) n) thm end


fun var_hyp_subst_tac no state =
    let val prms = Logic.strip_imp_prems(Logic.nth_prem(no, prop_of state));
        fun is_eq (Const (@{const_name HOL.eq}, _) $ (Var _) $ _) = true
           |is_eq (Const (@{const_name HOL.eq}, _) $ _ $ (Var _)) = true
           |is_eq _ = false
        val i = Library.find_index(fn (Const(@{const_name Trueprop},_)$X) => is_eq X)
                                  (prms)
        fun mks thm = if i >= 0
                      then ((rotate_tac i no) THEN etac @{thm thin_refl} no) thm
                      else no_tac thm
    in ((REPEAT mks)
        THEN
        (full_simp_tac (global_simpset_of @{theory Main}) no)) state
    end;


(* parallel tactic application, without any sanity checks *)

local exception FAILED of unit in

fun PARALLEL tac st =
  let
    fun try_tac g =
      (case SINGLE tac g of
        NONE => raise FAILED ()
      | SOME g' => g');
    val goals = Drule.strip_imp_prems (Thm.cprop_of st);
    val results = Par_List.map (try_tac o Goal.init) goals;
  in ALLGOALS (fn i => Goal.retrofit i 1 (nth results (i - 1))) st end
  handle FAILED () => Seq.empty;

end;


(* *********************************************************************** *)
(*                                                                         *)
(*  special tactics for TestGen                                            *)
(*                                                                         *)
(* *********************************************************************** *)

(*PRE : Must not be THYP or PO *)
fun uniformityI_tac ctxt prog = SELECT_GOAL
    let
        val thy = ProofContext.theory_of ctxt;
        val cert = Thm.cterm_of thy;
        val certT = Thm.ctyp_of thy;

        val core_tac =
          SUBGOAL (fn (term, i) =>
            let fun conv_All (Const (@{const_name All}, ty) $ Abs(n, ty', tt)) =
                      (Const (@{const_name Ex}, ty) $ Abs (n, ty', conv_All tt))
                  |conv_All(Const (@{const_name HOL.implies}, ty) $ P $ Q) =
                      (Const (@{const_name HOL.conj}, [boolT,boolT]--->boolT) $ P $ conv_All Q)
                  | conv_All tt = tt;
                val tt = dest_Trueprop term;
                val thyp = mk_Trueprop (@{const THYP} $ (@{const HOL.implies} $ conv_All tt $ tt));
            in rtac (Thm.instantiate ([], [(@{cpat "PROP ?psi"}, cert thyp)]) cut_rl) i end);

        val x = @{cpat "?x::?'a"};
        val {T, ...} = rep_cterm x;
        val count = Unsynchronized.ref 0;
        fun new_var () = cert (Var (("?X" ^ Int.toString (Unsynchronized.inc count) ^ "X", 0), T));
        fun exI_tac i st = rtac (Thm.instantiate ([], [(x, new_var ())]) @{thm exI}) i st;
        fun check_tac' (t,i) = case t of
                                   @{term Trueprop} $ t' => check_tac' (t',i)
                                 | @{term HOL.conj} $ u $ _ => if exists_prog prog u then 
                                                                 no_tac 
                                                             else all_tac

        val check_tac = SUBGOAL check_tac'
        val intro_PO = (rtac @{thm PO_I} 1) THEN (check_tac 1) 
                                            THEN REPEAT_DETERM ((rtac @{thm PO_grow} 1) THEN (check_tac 1))

        val finish_tac = (intro_PO THEN (rtac @{thm "conjI"} 1) THEN (strip_tac 2)) ORELSE (strip_tac 1)
    in
        EVERY [mp_fy 1, all_ify ctxt prog 1,
               core_tac 1, etac @{thm THYP_app1} 1,
               (REPEAT_DETERM o exI_tac) 1,
               finish_tac]
    end


(* ************************************************************************* *)
(* data separation lemma generation ...                                      *)
(* ************************************************************************* *)

fun refine_data_separation_lemma changed ctxt bal prem_no cands thm =
(* - bal controls of the refinement is in balancing mode
     (only parameters whose occurrence in the term is
      has strictly less than maximum height)
     or not.
   - prem_no indicates the number of the premise
   - cands is the global list of candidates.
     In balancing mode, parameters in maximal depth
     position were eliminated.
   - PRE maxidx is 0 for all mvars
   - POST maxidx is 0 for all mvars

 *)
let
    val thy    = ProofContext.theory_of ctxt
    val prem   = Logic.nth_prem(prem_no, prop_of thm)
    val params = Logic.strip_params(prem)
    val cands' = if bal then
                    let val term   = HOLogic.dest_Trueprop(hd
                                            (Logic.strip_assums_hyp prem))
                        val (k,mp) = max_depth(Term.subst_bounds
                                            ((rev(map Free params)),term))
                    in filter(fn p => forall(fn p'=> not(p=p')) (mp)) cands end
                 else cands;

    val var_no = find_index (fn(s,t)=> exists(fn (name,ty)=>
                                             s = name andalso t = ty) cands')
                             params
    val (name,ty)  = if var_no > ~1 then List.nth(params,var_no)
                     else raise (ERROR("no param found"))
    val Type(tn,_) = ty;
    val m          = get_case_split_of_dt thy tn;
    val n          = m RSN (prem_no, thm);
    (* pre: all occurrences of ?indname t1 ... tn have the same
            arguments and types  and occur in a critical pattern :
             ?indname t1 ... tn = C x1 ... xn
        ...
       The Thm.adjust_maxidx_thm reflects the fact, that only one
       MVar y should have been left - the others are erased due to
       hypsubst internally
    *)
    fun find_mvar_pat indname goal_no thm =
        let val goal = Logic.get_goal thm goal_no
            val params = Logic.strip_params goal
            fun  cc(Const(@{const_name "==>"}, _)
                    $ (Const(@{const_name Trueprop}, _)
                       $ (Const(@{const_name HOL.eq},t) $ A $ B))
                    $ C) = (case strip_comb(subst_bounds(
                                     rev(map Free params), A)) of
                             (Var(s,_),S) =>(if s = indname
                                             then (map dest_Free S, head_of B)
                                             else cc(C))
                            | (_, _)  => cc C)
               | cc _ = error("find_mvar_pat: No pattern found")
        in cc (strip_all_body goal) end
    fun lifter name S =
        let fun ll k [] = if 0 <=k then Bound(k)
                          else error"find_mvar:impossible lifter"
               |ll k ((s,t)::R) =
                      if s = name then Abs(s,t, ll 0 R)
                      else Abs(s,t, ll (k+1) R)
        in ll (~(length S)) S end
    fun typ_of_lifter (name:string) S =
        let fun tll S [] = error"find_mvar:impossible type for lifter"
               |tll S ((s,t)::R) =
                       if name = s
                       then ((rev S)@(t::(map snd R)), t)
                       else tll (t::S) R
        in  (op --->)(tll [] S) end
    val (lifter_params, constr) = find_mvar_pat ("y",1) prem_no (prop_of n)
    (* ("y",1) is a fact from the invariant maxidx=0 (see below) *)
    val subst0 = lifter name lifter_params
    val ty_lift = fastype_of subst0
    val ty_constr = fastype_of constr
    val subst = cterm_of thy subst0
    val ty_lift_body = body_type ty_lift
    val ty_lift_constr = body_type ty_constr
    val tyenv = Sign.typ_match thy (ty_lift_constr,ty_lift_body) Vartab.empty
                handle Type.TYPE_MATCH => Vartab.empty
    (* Hack that works for cases, where non-recursive positions in a type
       were GROUND types (like color in RBT_tree). *)
    fun tysubs2ctyps(x,(s,t)) = (ctyp_of thy (TVar(x,s)),ctyp_of thy t)
    val tsubst = map tysubs2ctyps (Vartab.dest tyenv)
    val ()    = (changed:=true)
    val var = cterm_of thy (Var(("y",1),ty_lift))
    val result = (Drule.instantiate_normalize (tsubst, [(var,subst)]) n)
                 |$> (ALLGOALS(fn n => TRY(hyp_subst_tac n)))
                 |$> prune_params_tac
in  Thm.adjust_maxidx_thm 0 result
    (* establishes POST and makes this fact therefore invariant. *)
    (* More safe, but also more expensive variant: standard result. *)
end

(* This is the core procedure of the data_separation_lemma generation, and,
   in a way, the core of the test case generator itself.

   It proceeds as follows: in a given theorem thm (usually an exhaustion lemma
   itself), the premisses (having the form !!x1..xn. A x1 .. xn ==> B in
   general) were selected and "refined" in a "sweep". This means that in a
   sweep, for each premisse an xi is selected (which must have a data-type,
   say ti) and the exhaustion theorem of ti is forward resolved into this
   premisse; cleaning it up (by prems_hyp_subst_tac) also erases xi as
   parameter of this premisse.
   In a next sweep, another xj may be selected and so forth, up to the moment
   were no parameters remained in the premisse that existed before the initial
   sweep (this the list parameters is the "cand" (candidates) x1 .. xj for
   all premisses which precomputed.)

   Then, the candidate list may be recomputed and the process repeated up to
   n times. Since any group of sweeps erases one generation of parameters,
   the parameter n corresponds to "depth+1" of the data term occuring in the
   premisse increases, i.e. from:

   [| ?y = Leaf ==> ?P;
      !!a tree1 tree2. ?y = Node a tree1 tree2 ==> ?P |]
   ==> ?P

   we get:

   [| ?y = Leaf ==> ?P; !!a. ?y = Node a Leaf Leaf ==> ?P;
      !!a aa tree1 tree2a. ?y = Node a Leaf (Node aa tree1 tree2a) ==> ?P;
      !!a aa tree1a tree2a. ?y = Node a (Node aa tree1a tree2a) Leaf ==> ?P;
      !!a aa tree1a tree2a ab tree1 tree2b.
            ?y = Node a (Node aa tree1a tree2a) (Node ab tree1 tree2b) ==> ?P
   |] ==> ?P

   This picture is slightly complicated by the fact that Isabelle does not
   change the parameters, but may reintroduce parameter erased by a previous
   sweep during the forward resolution step. Thus, the procedure may loop.

   Instead of not erasing the parameters during lemma-refinement (this
   leads to very large liftings of the meta-variables over these
   parameter lists and is thus expensive), we chose to use
   a particular test for "maximal" parameters in order to avoid this problem.

 *)
fun nrefine_data_separation_lemma changed ctxt 0 thm = thm
  | nrefine_data_separation_lemma changed ctxt n thm =
    let
      val thy = ProofContext.theory_of ctxt
      val isaDataType   = fn(_,ty)=> isDataType thy ty
      val thm_params    = maps Logic.strip_params (prems_of thm)
      val cands         = distinct (op=) (filter isaDataType thm_params)
      fun rds b(prno,t) = refine_data_separation_lemma changed ctxt b prno cands t
                          handle _ => t  (* FIXME avoid handle _ *)
      val ()            = (changed:=false)
      fun sweep b th    = List.foldr (rds b) th (1 upto (length(prems_of th)))
      val thm1          = sweep false thm       (* debalancing phase *)
      fun repeat_till_stable thm = let val ()   = (changed:=false)
                                     val thm' = sweep true thm
                                   in  if !changed
                                       then repeat_till_stable thm'
                                       else thm'
                                   end
    in  nrefine_data_separation_lemma changed ctxt
          (n-1)
           (if  !changed
            then thm1                     (* debalancing phase had no effect *)
            else repeat_till_stable thm1) (* balancing phase *)
    end;

(* NOTE: Currently, this core routine is restricted to
   "real free" variables; parameters will not be
   considered as a simple means to avoid looping ... *)
fun data_separation_tac ctxt var no state =
    let
        val thy    = ProofContext.theory_of ctxt
        val depth  = Config.get ctxt depth
	val (na,Type(tn,_)) = var
        fun ds_thm tn = nrefine_data_separation_lemma (Unsynchronized.ref false) 
						      ctxt 
						      depth 
						      (get_case_split_of_dt thy tn)
 
	val data_sep_thm = Drule.export_without_context (ds_thm tn)
        val brchs = length(prems_of data_sep_thm)
    in       
	((res_inst_tac ctxt [(("y",0),na)] data_sep_thm no)
             THEN (res_inst_tac ctxt 
				[(("x",0),na), (("k",0),Int.toString depth)]@{thm THYP_E_reg}
										 (no + brchs - 1))
             THEN (thin_tac ctxt (na^"= ?X") (no+brchs))) state
    end;

(* Copied from Isabelle2011 old_term.ML sources, for providing old variable order *)
(*Accumulates the Frees in the term, suppressing duplicates.*)
fun add_term_frees (t, frees: term list) = case t of
    Free   _ => Ord_List.insert Term_Ord.term_ord t frees
  | Abs (_,_,body) => add_term_frees(body,frees)
  | f$t =>  add_term_frees (f, add_term_frees(t, frees))
  | _ => frees;

fun term_frees t = add_term_frees(t,[]);

(* Included for complete backward compatibility *)
fun data_separation_old_tac ctxt depth no state =
     let
         val thy    = ProofContext.theory_of ctxt
        val prem   = Logic.nth_prem(no, prop_of state)
        val params = Logic.strip_params(prem)
        val term   = HOLogic.dest_Trueprop(Logic.strip_assums_concl prem)
        val frees  = map dest_Free (term_frees term)
(*      val cands  = filter (fn (_,ty) => isDataType thy ty) (params@frees) *)
        val cands  = filter (fn (_,ty) => isDataType thy ty) (frees)
        fun ds_thm tn =
          nrefine_data_separation_lemma (Unsynchronized.ref false) ctxt depth (get_case_split_of_dt thy tn)
    in (case cands of
          [] => no_tac
        | (na,Type(tn,_))::_ =>
                    let val data_sep_thm = Drule.export_without_context (ds_thm tn)
                        val brchs = length(prems_of data_sep_thm)
                    in       (res_inst_tac ctxt [(("y",0),na)] data_sep_thm no)
                        THEN (res_inst_tac ctxt [(("x",0),na),
                                                   (("k",0),Int.toString depth)]
                                                  @{thm THYP_E_reg}
                                                  (no + brchs - 1))
                        THEN (thin_tac ctxt (na^"= ?X") (no+brchs))
                    end)
       (state)
end


(* ************************************************************************* *)
(* normal form computation ...                                               *)
(* ************************************************************************* *)

fun to_TNF ctxt prog term no state =
    let
        val thy     = ProofContext.theory_of ctxt;
        val prems   = Logic.prems_of_goal term no
        val concl   = Logic.concl_of_goal term no

        val prems'  = filter (exists_prog prog) prems;
        val concl'  = filter (exists_prog prog) [concl];
        fun subst t = (Thm.cterm_of thy (Var(("P",0),HOLogic.boolT)),
                       Thm.cterm_of thy t);
        (*  swap_tac : tactic * term -> tactic; takes one term containing
            "prog" and swap it into the conclusion as disjoint *)
        val swap_accumulate =  @{lemma "[| P ; ~P | Q |] ==> Q" by auto}
        fun swap_tac (tm,tc) = tc THEN
                               (TacticPatch.eres_terminst_tac
                                  []
                                  [subst (HOLogic.dest_Trueprop tm)]
                                  swap_accumulate
                                  no)
        (* rev_tac swaps term "a" into conclusion *)
        fun rev_tac thy a = TacticPatch.eres_terminst_tac
                                  []
                                  [subst (HOLogic.dest_Trueprop a)]
                                  @{thm rev_notE}
                                  no
    in ((case (prems',concl') of
          ([],[])   => error ("Subgoal " ^Int.toString(no)
                             ^" can not be converted into TNF.")
          (* case that conclusion does not contain "prog", but several
             assumptions do: pick one of them, revert first *)
         |(a::R,[]) =>(rev_tac thy a
                       THEN
                       (List.foldl swap_tac all_tac R))
         |(R,_)     => (List.foldl swap_tac all_tac R)
        )
        THEN
         (simp_tac HOL_ss no)) state
    end;


(* TODO: adb *)
(*
fun normalize_TNF prog top =
    let val prems = prems_of top
        val mrs   = prems ~~ (1 upto (length prems))
    in  List.foldl (fn(arg,t)=> t THEN to_TNF prog arg) all_tac mrs (top)
    end
 *)
fun normalize_TNF ctxt prog top =
    let val mrs   = (length (prems_of top)) downto 1
    in  List.foldl (fn (arg,t) => t THEN to_TNF ctxt prog (prop_of top) arg)
              all_tac
              mrs
              top
    end


(* TODO: adb *)
fun minimize_TNF ctxt = distinct_subgoals_tac
(* This should do:
   - sorting assumptions in all subgoals
   - sorting parameters of subgoals
   - sortieren der clauses -
     weniger asms zuerst (defer_tac)
   - eliminieren von assumptions subsumed in other clauses
     [| P;R |] ==> A;
     [| P;Q;R  |] ==> A
     ------------------
     [| P;R |] ==> A;
     [| P;R  |] ==> A
     durch thin_tac
   - eliminieren von subsumed or's,
     a particular weakness of our procedure so far.
     [| P;R |] ==> A;
     [| ~P;Q |] ==> B
     [| R; Q|] ==> A|B;
     ------------------
     [| P;R |] ==> A;
     [| ~P;Q |] ==> B
    - finally distinct_subgoals_tac
*)


(* orig version *)
(*
fun gen_test_case_tac thy clasimp depth breadth prog  =
    EVERY[ALLGOALS (asm_full_simp_tac (snd clasimp)),
          safe_tac ((global_claset_of thy) addss (global_simpset_of thy)),
          normalize_TNF thy prog,
          minimize_TNF,
          ALLGOALS (fn n => TRY(uniformityI_tac prog n))];
*)



(* ************************************************************************* *)
(* test case generation ...                                                  *)
(* ************************************************************************* *)

fun SORT ctxt rules n thm = let
    val matches = maps (fn rule => rule ctxt n thm) rules
    val sorted = map (fn (i,tac) => tac) (sort (fn (x,y) => rev_order (int_ord (fst x, fst y))) matches) 
in
    (EVERY sorted) thm
end
			   
fun testenv_tac f ctxt = let 
    val testenv = TestEnv.get_data ctxt
in
    f testenv ctxt
end

fun EXEC ctxt tac str = 
    (if Config.get ctxt profiling then start_clock_tac str else all_tac)
        THEN tac
	THEN (if Config.get ctxt trace then print_tac ("\n:"^str^"\n") else all_tac)
	THEN (if Config.get ctxt profiling then stop_clock_tac str else all_tac)

fun finalize_tac ctxt prog = EXEC ctxt (ALLCASES(TRY'(uniformityI_tac ctxt prog))) "main uniformity NF"
			 
fun gen_test_case_tac ctxt prog = let
    val breadth = Config.get ctxt breadth
    val testenv = TestEnv.get_data ctxt
    val rules = TestEnv.get_test_derivation_rules testenv

    val begin_profiling_tac = if Config.get ctxt profiling then
                                  start_clock_tac "unnamed_test_thm" THEN start_clock_tac "gen_test_cases"
                              else
                                  all_tac
				  
    val end_profiling_tac = if Config.get ctxt profiling then
                                stop_clock_tac "gen_test_cases" THEN stop_clock_tac "unnamed_test_thm"
                            else
                                all_tac

    val pre_normalize_TNF_tac = testenv_tac get_pre_normalize_TNF_tac

    val sort_tac = SELECT_GOAL (SORT ctxt rules 1)
in  
    EVERY[begin_profiling_tac,
	  EXEC ctxt (ALLCASES((asm_full_simp_tac (simpset_of ctxt)))) "pre-simplification",
	  EXEC ctxt (REPEAT_DETERM_N breadth (ALLCASES sort_tac)) "main completed",
          EXEC ctxt (pre_normalize_TNF_tac ctxt) "pre_norm",
          EXEC ctxt (normalize_TNF ctxt prog) "TNF",
	  if Config.get ctxt no_finalize then all_tac else finalize_tac ctxt prog,
	  end_profiling_tac]
end 

fun data_separation_rule ctxt no state = let
    val thy    = ProofContext.theory_of ctxt
    val prem   = Logic.nth_prem(no, prop_of state)
    val params = Logic.strip_params(prem)
    val term   = HOLogic.dest_Trueprop(Logic.strip_assums_concl prem)
    val frees  = Term.add_frees term []
		     (*      val cands  = filter (fn (_,ty) => isDataType thy ty) (params@frees) *)
    val cands  = filter (fn (_,ty) => isDataType thy ty) (frees)
    fun tac cand = TRY (CHANGED_PROP (ALLCASES (data_separation_tac ctxt cand)) 
                                     THEN (ALLCASES (TRY' hyp_subst_tac)))
                       THEN (if Config.get ctxt trace then print_tac "After data separation" else all_tac)
in
    map (fn cand => (100, tac cand)) cands
end

fun to_HCNF_rule ctxt no state = 
    [(50, EXEC ctxt (testenv_tac get_pre_safe_tac ctxt) "Simp"
	       THEN EXEC ctxt (TRY (safe_tac ctxt)) "HCN")]

fun minimize_rule ctxt no state = 
    [(10, EXEC ctxt (testenv_tac get_pre_minimize_TNF_tac ctxt) "pre_minimize"
	       THEN EXEC ctxt (minimize_TNF ctxt) "MinimTNF"
               THEN (ALLCASES(asm_full_simp_tac (simpset_of ctxt)))
               THEN (TRY (safe_tac ctxt)))] 

(* Included for complete backward compatibility *)
fun gen_test_case_old_tac ctxt prog  =
    let val testenv = TestEnv.get_data ctxt
        val depth = Config.get ctxt depth
        val breadth = Config.get ctxt breadth

        val prep_HCNF = get_pre_safe_tac testenv
        val to_HCNF = TRY (safe_tac ctxt)
        val pre_normalize_TNF_tac = get_pre_normalize_TNF_tac testenv
        val pre_minimize_TNF_tac = get_pre_minimize_TNF_tac testenv
        fun trace_tac b str = (if Config.get ctxt trace then print_tac ("\n"^b^":"^str^"\n")
                               else all_tac)
        fun exec phase b str = (if Config.get ctxt profiling then start_clock_tac str else all_tac)
                               THEN phase THEN trace_tac b str THEN
                               (if Config.get ctxt profiling then stop_clock_tac str else all_tac)

        fun standard k b =
            EVERY[if Config.get ctxt trace then print_tac "Enter" else all_tac,
                  exec (prep_HCNF ctxt) b "Simp",
                  exec (to_HCNF) b "HCN",
                  exec (pre_normalize_TNF_tac ctxt) b "pre_norm",
                  exec (normalize_TNF ctxt prog) b "TNF",
                  exec (pre_minimize_TNF_tac ctxt) b "pre_minimize",
                  exec (minimize_TNF ctxt) b "MinimTNF"]
        val dst = data_separation_old_tac ctxt depth


        (*True if the two theorems have the same prop field, ignoring hyps, der, etc.*)
        val eq_thm_prop = op aconv o pairself Thm.full_prop_of;

        fun main k st =(if k < breadth (* not yet all frees per goal
                                          less breadth exhausted*)
                        then   (ALLCASES(fn n => TRY(dst n) )) THEN
                               (* try exhaustion per goal *)
                               (COND(fn st' => not (eq_thm_prop(st, st')))
                                        (* something 's changed *)
                                        (     (ALLCASES (fn n => TRY(hyp_subst_tac n) ) ) (* REALLY ? *)
                                         THEN (standard k "1")
                                         THEN (main (k+1))
                                        )
                                        (standard k "2"))
                        else standard  k "0" (* run out of breadth *)) (st)
        val begin_profiling_tac = if Config.get ctxt profiling then
                                      start_clock_tac "unnamed_test_thm" THEN start_clock_tac "gen_test_cases"
                                  else
                                      all_tac

        val end_profiling_tac = if Config.get ctxt profiling then
                                      stop_clock_tac "gen_test_cases" THEN stop_clock_tac "unnamed_test_thm"
                                  else
                                      all_tac
    in  EVERY[begin_profiling_tac,
              exec (ALLCASES((asm_full_simp_tac (simpset_of ctxt)))) "" "pre-simplification",
              exec (main 0) "" "main completed",
              exec (ALLCASES(TRY'(uniformityI_tac ctxt prog))) "" "main uniformity NF",
              end_profiling_tac]
    end;


(* ************************************************************************* *)
(*                                                                           *)
(* testdata generation        ...                                            *)
(*                                                                           *)
(* ************************************************************************* *)

  (* *********************************************************************** *)
  (* Random Solver                                                           *)
  (* *********************************************************************** *)

fun random_insts thy cod_tab vars () = map(fn(x as Var(s,t))=>
                                             (cterm_of thy x,
                                              cterm_of thy (random_term thy cod_tab t))) vars
    
fun single_rand_inst_tac ctxt vars thm = let
    val thy = ProofContext.theory_of ctxt
    val te = TestEnv.get_data ctxt
    val cod_tab = #cod_term_tab(rep_testenv te)
    val insts = random_insts thy cod_tab vars ()
in
    Seq.single (Drule.instantiate_normalize ([], insts) thm)
end

fun premvars n thm = let
    val prem = Logic.nth_prem(n, prop_of thm)
in
    map Var (Term.add_vars prem [])
end

fun solve_by_simp_tac ctxt = SOLVED' (full_simp_tac (simpset_of ctxt))

fun random_inst_tac ctxt iters n thm = let
    val _    = if iters > 0 then tracing ("Random solving subgoal "^Int.toString(n)) else () 
    val single_tac = (single_rand_inst_tac ctxt (premvars n thm)) THEN (solve_by_simp_tac ctxt n)
in
    (FIRST (replicate iters single_tac)) thm
end

fun list_abs_var t [] = t
  | list_abs_var t ((x as Var(_, T))::vars) = Abs(Name.internal Name.uu, T, abstract_over(x, list_abs_var t vars))

fun iterate f 0 = NONE
  | iterate f k = case f () handle Match => NONE
                   of NONE => iterate f (k - 1) | SOME q => SOME q;

fun certify_pairs thy l = map (fn (x,y) => (cterm_of thy x, cterm_of thy y)) l

fun uncertify_pairs l = map (fn (x,y) => (term_of x, term_of y)) l

fun quickcheck_tac' ctxt iters n thm =
  let
      val size = 15
      val thy = ProofContext.theory_of ctxt
      val vars = premvars n thm
      val prem = Logic.nth_prem(n, prop_of thm)
      val neg =  @{term Not} $ (HOLogic.dest_Trueprop prem)
      val neg' = list_abs_var neg vars
      val tester = Codegen.test_term ctxt neg'
      fun with_tester k = iterate (fn () => tester k) iters
      fun with_size k = if k > size then NONE
                        else
                            (case with_tester k
                              of NONE => with_size (k + 1)
                               | SOME q => SOME q);
  in case with_size 1 of
         SOME insts => let
             val instantiated = Drule.instantiate_normalize ([], certify_pairs thy (vars ~~ insts)) thm
         in
             full_simp_tac (simpset_of ctxt) n instantiated
         end
       | NONE => Seq.empty
  end

fun quickcheck_tac ctxt iters n thm = let
    val tac = Object_Logic.full_atomize_tac THEN' (quickcheck_tac' ctxt iters)
in 
    (case (Seq.pull (tac n thm)) of 
         SOME (x, xq) => Seq.single x
       | NONE => Seq.empty)
    handle ERROR _ => Seq.empty (* Catch "unable to generate code" exceptions *) 
end

fun pretty_insts ctxt msg insts = let
    fun mkeq (x,y) = Const(@{const_name HOL.eq}, dummyT) $ x $ y
    val pretty_insts = map (fn inst => (Syntax.pretty_term ctxt (mkeq inst))) insts
in
    Pretty.string_of (Pretty.big_list (msg^":") pretty_insts)
end

fun scan_def (Const (@{const_name HOL.eq}, _) $ t $ u) = (t,u)
  | scan_def _ = error ("Unexpected SMT counterexample format")

fun mkinsts thy vars ce = let
    val defs = map scan_def ce
    (* val _ = tracing (pretty_insts @{context} "Raw SMT counterexample" defs) *)
    val defs' = defs @ (map swap defs)
    fun is_Free_or_Var t = (is_Free t) orelse (is_Var t)
    val real_defs = filter (fn (x,_) => is_Free_or_Var x) defs'
    val var_insts = map (fn x => (x, subst_atomic real_defs x)) vars
    (* val _ = tracing (pretty_insts @{context} "Generated instantiation from SMT counterexample" var_insts) *)
in
    certify_pairs thy var_insts
end

(* a wrapper around smt_tac' that fixes the provided counterexamples *)
fun smt_ce_tac ctxt rules =     
    Subgoal.FOCUS(fn {context, prems, schematics, ...} => (fn thm => 
let
    val prof_name = "SMT"
    val start_tac = (if Config.get ctxt profiling then start_clock_tac prof_name else all_tac)
    val stop_tac = (if Config.get ctxt profiling then stop_clock_tac prof_name else all_tac)
in
    (start_tac THEN (SMT_Solver.smt_tac' context (rules @ prems) 1) THEN stop_tac) thm
                                                                                   |> Seq.hd |> Seq.single
                                                                                                      
    handle SMT_Failure.SMT (SMT_Failure.Counterexample {is_real_cex = is_real_cex', 
                                                        free_constraints, 
                                                        const_defs}) => 
           let
               val _ = if Config.get ctxt profiling then stop_clock prof_name else ()
               val inv_insts = map swap (snd schematics)
               val term_insts = uncertify_pairs inv_insts
               val free_constraints' = map (subst_atomic term_insts) free_constraints
               val const_defs' = map (subst_atomic term_insts) const_defs
           in
               raise SMT_Failure.SMT 
                         (SMT_Failure.Counterexample {is_real_cex = is_real_cex', 
                                                      free_constraints = free_constraints', 
                                                      const_defs = const_defs'})
           end

         | exc => let
               val _ = if Config.get ctxt profiling then stop_clock prof_name else ()
           in
               raise exc
           end

end)) ctxt

fun try_inst_tac ctxt insts n thm = let
    val thm' = Drule.instantiate_normalize ([], insts) thm
in
    solve_by_simp_tac ctxt n thm'
end
    
fun smt_inst_tac ctxt rules n thm = let
    val thy = ProofContext.theory_of ctxt
    val prem = Logic.nth_prem(n, prop_of thm)
    val neg = @{const "==>"}
                    $ prem
                    $ (@{const "Trueprop"} $ @{const "False"})
    val goal = Goal.init (cterm_of thy neg) 
in
    (Seq.hd (smt_ce_tac ctxt rules 1 goal); Seq.empty)
    handle SMT_Failure.SMT (SMT_Failure.Counterexample {free_constraints, const_defs, ...}) 
           => try_inst_tac ctxt (mkinsts thy (premvars n thm) (free_constraints @ const_defs)) n thm
         | SMT_Failure.SMT _ => Seq.empty
end

fun add_bound_tac ctxt bound t =
    Subgoal.FOCUS_PARAMS(fn {context, schematics, ...} => let
                                val thy = ProofContext.theory_of context
                                val ct = Thm.instantiate_cterm schematics (cterm_of thy t)
                                val bound_term = cterm_of thy (HOLogic.mk_nat bound)
                                val xtype = ctyp_of thy (TVar (("'a",0), @{sort "Nat.size"}))
                                val tinst = (xtype, ctyp_of_term ct)
                                val k = cterm_of thy (Var (("k",0), @{typ "nat"}))
                                val x = cterm_of thy (Var (("x",0), type_of t))
                                val inst1 = (k, bound_term)
                                val inst2 = (x, ct)
                            in
                                TacticPatch.res_terminst_tac [tinst] [inst1, inst2] @{thm add_bound} 1
                            end) ctxt

fun add_bounds_tac ctxt bound ts = let
    fun next (t, tac) = tac THEN' (add_bound_tac ctxt bound t)
in
    List.foldl next (K all_tac) ts
end

fun bounded_smt_tac ctxt bound rules =
    Subgoal.FOCUS_PARAMS(fn {context, ...} => 
                            (fn thm => let
                                    val thy = ProofContext.theory_of context
                                    val datatype_vars = filter (fn x => isDataType thy (type_of x)) (premvars 1 thm)
                                in
                                    ((add_bounds_tac context bound datatype_vars 1)
                                         THEN (smt_inst_tac context rules 1)
                                         THEN (ALLGOALS (full_simp_tac (simpset_of context)))) thm
                                end)) ctxt

fun abs_data_tac ctxt atdata = let
    fun single_abs_data_tac thm = 
        Subgoal.FOCUS_PARAMS(fn {context, ...} => SOLVE (Method.insert_tac [thm] 1 THEN auto_tac context)) ctxt

in
    FIRST' (map single_abs_data_tac atdata)
end

fun trig_thm thy thm = let
    val is_trig = Pattern.matches_subterm thy (@{const SMT.trigger}, prop_of thm)
                  orelse not (exists_subterm is_Var (prop_of thm))
in
    if is_trig then
        thm
    else
        thm RS @{thm "trigger_intr"} handle THM _ => thm
end

fun test_solve_tac ctxt atdata = let
    val thy = ProofContext.theory_of ctxt
    val te = TestEnv.get_data ctxt
    val remove_po = EqSubst.eqsubst_tac ctxt [0] [@{thm PO_def}]
    val total_iterations = Config.get ctxt iterations
    val max_simple_iterations = 50
    val simple_iterations = Int.min(total_iterations, max_simple_iterations)
    val further_iterations = Int.max(total_iterations - max_simple_iterations,0)
    val smt_facts = map (trig_thm thy) (TestEnv.get_smt_facts ctxt)

    val testgen_smt_tac = if Config.get ctxt SMT then
                              bounded_smt_tac ctxt (Config.get ctxt depth) smt_facts
                          else
                              K no_tac                  
in
    remove_po THEN' (FIRST'[abs_data_tac ctxt atdata,
                            random_inst_tac ctxt simple_iterations,
                            quickcheck_tac ctxt further_iterations,
                            testgen_smt_tac])
end

fun is_solved n thm = let
    fun is_unsolved_po i = 
        not (null (inter (op =) (premvars n thm) (premvars i thm))) 
            andalso i <> n
in
    not (exists is_unsolved_po (1 upto (nprems_of thm)))
end

(* Instantiate the remaining variables in test cases
with random terms *)

fun finalize_test_tac ctxt n thm = let
in
    (COND (is_solved n) (single_rand_inst_tac ctxt (premvars n thm)) all_tac) thm
end

fun solve_all ctxt atdata state =
    state |$> PARALLEL (ONLY_POS (TRY' (test_solve_tac ctxt atdata)) 1)
          |$> ALLCASES (finalize_test_tac ctxt) (* cannot parallelize here *)



  (* *********************************************************************** *)
  (* Normalizer                                                              *)
  (* *********************************************************************** *)
  (* motivation: type instantiation may make
     constraints of test-cases unsolvable.
     Unsolvable cases should be removed before random-solving.
     Since constraints with Mvars were not considered
     by simptac and arith_tac, they were substituted
     against arbitrary free vars aforehand.

fun norm_tac ctxt n thm =
   let  val thy   = ProofContext.theory_of ctxt
        val m     = Thm.nprems_of thm
        val prem  = Logic.nth_prem(n, prop_of thm)
        val k     = length (Logic.strip_imp_prems(prem))
        val goal  = Thm.trivial(cterm_of thy prem)
        val tvars = OldTerm.term_vars prem
        val insts = map(fn(x as Var((s,a),t))=>
                            (cterm_of thy x,
                             cterm_of thy (Free("XXX"^s^(Int.toString a),t))))
                       (tvars)
        val cleanup    = List.foldr (fn (_,tac) => (atac (n)) THEN tac)
                               all_tac (1 upto k)
   in  (
        (rtac (instantiate ([],insts) goal) n)
        THEN
        (full_simp_tac (simpset_of ctxt) n)  (* may erase goal n.
                                 --> test case unsolvable. *)
        THEN
        (fn thm' => if (nprems_of thm' < m+k) (* goal erased *)
                    then cleanup thm'
                    else no_tac thm )(* report failure *)
       )(thm)
   end *)

fun gen_test_data name context =
    let val ctxt = Context.proof_of context
        val thy = ProofContext.theory_of ctxt
        val te = TestEnv.get_data ctxt
        val bound = Config.get ctxt type_range_bound
        val candTs= #type_candicates(rep_testenv te)
        val type_grounds = groundT_thm thy bound candTs
        val testthm  = (the(Symtab.lookup(thm_tab te)name)
                     handle Option.Option => error("No test theorem: "^name^" available!"))
        val atdata = (the(Symtab.lookup(absdata_tab te)name)
                      handle Option.Option => [])
        val jdmts = map (solve_all ctxt atdata) (type_grounds testthm)
        val te' =
            jdmt_tab_update
               (Symtab.update(name,jdmts)(jdmt_tab te))
               te
    in TestEnv.map_data (K te') context end;

fun select_goals pred thy name = let
    val te = TestEnv.get_data_global thy
in
    maps (convert_goals_to_metahyps pred)
             (the(Symtab.lookup(jdmt_tab te)name)
              handle Option.Option => error("No data statement: "^name^" available!"))
end

fun is_test_data t = (is_test t) andalso not (exists_subterm is_Var t)

val get_test_data = select_goals is_test_data

val get_test_hyps = select_goals is_thyp

val get_pos = select_goals is_po


(* *********************************************************************** *)
(*                                                                         *)
(* Initial test environment                                                *)
(*                                                                         *)
(* *********************************************************************** *)

val _ = Context.>> (TestEnv.map_data
  ((pre_safe_tac_update (fn ctxt => ALLCASES (asm_full_simp_tac (simpset_of ctxt))))
       #> (add_test_derivation_rule data_separation_rule)
	      #> (add_test_derivation_rule to_HCNF_rule)
		     #> (add_test_derivation_rule minimize_rule)));

val mt_testenv = get_data (ML_Context.the_local_context ());


(* ************************************************************************* *)
(* micellaneous functions ...                                                *)
(* ************************************************************************* *)


fun animate thy name =  (* HACK - very simplistic implementation *)
    let val ctxt = ProofContext.init_global thy |> Config.put simp_trace true
        fun animate_data_statement thm =
                (writeln ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>";
                 Pretty.writeln (Display.pretty_thm ctxt thm);
                 writeln "============================================";
   (*              asm_full_simplify (global_simpset_of(theory_of_thm(thm))) thm; *)
                 asm_full_simplify (simpset_of ctxt) thm;
                 writeln "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
    in
        List.app animate_data_statement (get_test_data thy name)
    end;

fun gen_testscript put smlput fname dataname thy =  let
    val te = TestEnv.get_data_global thy
	     
    val margin = 58
    val mode = []
    val put_term = Syntax.read_term_global thy put
    val put_is_const = is_Const put_term
    val long_put = if put_is_const
                   then let
                           val Const(s,_) = put_term
                       in
                           s
                       end
                   else put

    fun output_code gr  xs =
        let
            val module = ""
            val code = List.mapPartial (fn s =>
                                           let val c as (_, module', _) = Graph.get_node gr s
                                           in if module = "" orelse module = module' then SOME (s, c) else NONE end)
                                       (rev (Graph.all_preds gr xs));
        in
            implode (map (#3 o snd) code)
        end;


    fun gen_generate_tc_code prep_term num thy  (speclist: (term list * term) list)  =  let
        val numstr = Int.toString(num)
        fun image_type typ = (case typ of
                                  Type(s,[t]) =>  Type(s,[t])
                                | Type(@{type_name fun},(t::tl)) => image_type (hd (rev tl))
                                | t => t);
        val emptygr : Codegen.codegr = (Graph.empty,
                                ((Symtab.empty, Symtab.empty), (Symtab.empty, Symtab.empty)));

        val gen_wrapper = Config.get_global thy wrapper;
        val gen_toString = Config.get_global thy toString;
        val initGraph = Codegen.new_node ("<Top>", (NONE, "", "")) emptygr
        val gr = initGraph
        val smlput = if (smlput ="") then put else smlput

        fun generate (speclist: (term list * term) list) = let
            fun arity_of_type typ = (case typ of
                                         Type(s,[]) =>  0
                                       | Type(@{type_name fun},(t::tl)) => 1 + arity_of_type (hd tl)
                                       | _ => 0
                                    )

            fun arity_of src t =
                case t of
                    (Const(s,T)) => if (s=src) then (arity_of_type T) else 0
                  | (Free(s,T)) => if (s=src) then (arity_of_type T) else 0
                  | (Var(i,T)) => 0
                  | (Bound(i)) => 0
                  | (Abs(a,T,t)) => arity_of src t
                  | (f$t) => Int.max ((arity_of src f),(arity_of src t))

            val put_arity = if (gen_wrapper)
                            then List.foldl Int.max 0 (map (arity_of long_put)
                                                      (maps(fn(xs,y)=>y::xs) speclist))
                            else 0
            fun wrap src t =
                case t of
                  (Const(@{const_name Char},T)$x$y) =>  (* Crude Hack for converting char literals *)
                      (Free("TestHarness.string_to_char",(Type (@{type_name char}, [])-->(Type (@{type_name char}, [])))))
                                                      $(Const(@{const_name Char},T)$(wrap src x)$(wrap src y))
                  | (Const(s,T))
                    => if (s=src)
                       then if(gen_wrapper)
                            then Free("eval",T)
                            else Free(smlput,T)
                       else Const(s,T)
                  | (Free(s,T))
                    =>  if (s=src)
                        then if(gen_wrapper)
                             then Free("eval",T)
                             else Free(smlput,T)
                        else Free(s,T)
                  | (Var(i,T)) => Var(i,T)
                  | (Bound(i)) => Bound(i)
                  | (Abs(a,T,t)) => Abs(a,T,wrap (src) t)
                  | (f$t) => (wrap (src) f )$ (wrap (src) t)
                  | t => t

            fun gencode (gr, (t:term)) = swap (Codegen.invoke_codegen thy mode (Codegen.mk_deftab thy) "<Top>" ""
                                                                      false (wrap long_put t) gr)

          fun to_code gr termlist = Library.foldl_map (gencode)  (gr, map (prep_term thy) termlist);
          fun codeToString clist = (map (fn p => Pretty.string_of_margin margin
                                                     (Pretty.block [Pretty.brk 1, p])) clist);
            fun code_spec (gr,(pre,post)) = let
                val (gr_pre, ps_pre) = to_code gr pre;
                val ps_pre = codeToString ps_pre
                val (gr_post,ps_post) = to_code gr_pre [post];
                val ps_post = codeToString ps_post

            in
                (gr_post, (ps_pre, hd ps_post))
            end

            val (gr', code) = Library.foldl_map (code_spec) (gr, speclist)

            fun check_code (num,(pre,post)) = let
                val numstr = Int.toString num;
                val num = num +1;
            in
                (num, "val _ = print (\"\\nRunning Test Case "^numstr^":\\n\")\n"
                      ^"val pre_"^numstr
                      ^"  = [" ^
                      (space_implode ",\n             "
                                     (map (fn s => "fn () => (" ^ s ^ ")") pre)
                      )
                      ^ "];\n" ^
                      "val post_"^numstr^" = " ^
                      "fn () => (" ^ post ^ ")"
                      ^ ";\n"^
                      "val res_"^numstr^"  = TestHarness.check retval pre_"^numstr^
                      " post_"^numstr^";\n"^
                      "val testres = testres@[res_"^numstr^"];\n\n"
                )
            end

            val prelude_str = if Config.get_global thy prelude
                              then   (output_code (fst gr')   ["<Top>"] )
                              else ""
            val (num,check_str)   = Library.foldl_map (check_code) (num,code);
            fun toString typ = (case typ of
                                    Type ("boolN", [])         => "Bool.toString"  (* FIXME "boolN" !?? *)
                                  | Type (@{type_name int}, [])  => "Int.toString"
                                  | Type(@{type_name real}, []) =>" Real.toString"
                                  | Type(@{type_name list},[T]) =>
                                    "(fn l => (\"[\"^(List.foldr (fn (s1,s2) \n"^
                                    "                 => (if s2 = \"\" then s1 else s1^\", \"^s2))\n"^
                                    "                    \"\" (map (fn x => "^(toString T)^
                                    " x) l))^\"]\")) "
                                  | _ => "(fn s => \"\") ")
            fun smltype_of typ = (case typ of
                                    Type ("boolN", [])       => SOME "bool"    (* FIXME "boolN" !?? *)
                                  | Type (@{type_name int}, [])  => SOME "int"
                                  | Type(@{type_name real}, []) => SOME "real"
                                  | Type(@{type_name list},[T]) => (case smltype_of T of
                                                                   SOME t => SOME ("("^t^" list)")
                                                                 | NONE => NONE)
                                  | _ => NONE
                              )
            val wrap_str = (if (gen_wrapper)
                           then let
                                   val cdtab = #cod_term_tab(rep_testenv te)
                                   val putT = if (put_is_const)
                                              then image_type (type_of (Syntax.read_term_global thy put))
                                              else
                                                  let
                                                      fun isput (s,t) = if (s=put)
                                                                            then true
                                                                            else false
                                                        | isput _ = false
                                                      fun frees (a,b) = Term.add_frees b []
                                                      val postF = filter isput (maps frees speclist)
                                                      val (_,freetype) = hd postF
                                                  in
                                                      image_type freetype
                                                  end

                                   val retval_term = random_term thy cdtab putT
                                   val (gr,ps) = to_code initGraph (* (Graph.new_node ("<Top>", (NONE, ""))
                                                                         Graph.empty) *) [retval_term];
                                   val retval = hd (codeToString ps);


                                   fun vars n =  if(n = 0) then ""
                                                 else " x"^Int.toString(n)^" "^(vars (n-1))
                                   val toStrings = if(gen_toString = "")
                                                   then "fun toString  a = " ^(toString putT)^" a;"
                                                   else "";

                                   val retvaltype = case (smltype_of putT) of
                                                      SOME t => ":("^t^")"
                                                    | NONE => ""
                               in
                                   "val return = ref ("^retval^retvaltype^");\n"^
                                   "fun eval "^(vars (put_arity))^" =  let val ret = "^smlput^" "
                                   ^(vars (put_arity))^" in ((return := ret);ret)end\n"  ^
                                   "fun retval () = SOME(!return);\n" ^
                                   toStrings^"\n\n"
                               end
                           else
                               let
                                   val toStrings = if(gen_toString = "")
                                                   then  "fun toString a = \"\";"
                                                   else "";
                               in
                                   "fun retval () = NONE;\n"^
                                   toStrings^"\n\n"
                               end)

                val toStringName = if (gen_toString = "") then "toString" else gen_toString;

        in

            "(*******************************************************************\n"
            ^" *                         Test-Driver \n"
                        ^" *           generated by HOL-TestGen "^testgen_version^"\n"
            ^" *******************************************************************)\n\n"
            ^"structure TestDriver : sig end = struct\n\n"
            ^ Config.get_global thy setup_code ^"\n\n"
            ^prelude_str
            ^ Config.get_global thy dataconv_code ^"\n\n"
            ^ wrap_str
            ^ "val testres = [];\n\n"
            ^ (List.foldl (fn (s1,s2) => s1^s2) "" check_str)
            ^ "val _ = TestHarness.printList "^toStringName^" testres;\n\n"
            ^ "end\n"
        end
    in
        Print_Mode.setmp [] generate speclist
    end;


    val generate_tc_code_i =  gen_generate_tc_code (K I);


    val td = get_test_data thy dataname;
    val prop_terms = map (fn p => let val _ $ ret = concl_of p  in ret end) td;
    val prems_terms = map (fn data => (
                              map (fn p => let val _ $ ret = p in ret end) (prems_of data))) td;

    val spec = map2 (fn x => fn y => (x,y))  prems_terms prop_terms
    val tc   = generate_tc_code_i 0 thy spec;
    val _    = File.write (Path.explode fname) tc;
in
    thy
end;


(* module setup *)

val setup =
  trace_setup #>
  profiling_setup #>
  no_finalize_setup;

end (*struct *);

